import numpy as np
import pandas as pd
import pytest

from wittgenstein import preprocess

DF = pd.read_csv("house-votes-84.csv").head(10)
CLASS_FEAT = "Party"
DEFAULT_CLASS_FEAT = "Class"
POS_CLASS = "democrat"
SPLIT_SIZE = 0.6

X_DF = DF.drop(CLASS_FEAT, axis=1)
Y_DF = DF[CLASS_FEAT]

XY_NP = DF.values
X_NP = X_DF.values
Y_NP = Y_DF.values
NP_CLASS_FEAT = 0


def test_preprocess_training_data_Xy_df():
    preprocess_params = {
        "trainset": DF,
        "y": None,
        "class_feat": CLASS_FEAT,
        "pos_class": POS_CLASS,
        "feature_names": None,
        "n_discretize_bins": None,
        "verbosity": 0,
    }
    (
        df,
        ret_class_feat,
        ret_pos_class,
        ret_bin_transformer_,
    ) = preprocess.preprocess_training_data(preprocess_params)
    assert ret_class_feat == CLASS_FEAT
    assert ret_pos_class == POS_CLASS
    assert ret_bin_transformer_ == None


def test_preprocess_training_data_X_df():
    preprocess_params = {
        "trainset": X_DF,
        "y": Y_DF,
        "class_feat": CLASS_FEAT,
        "pos_class": POS_CLASS,
        "feature_names": None,
        "n_discretize_bins": None,
        "verbosity": 0,
    }
    (
        df,
        ret_class_feat,
        ret_pos_class,
        ret_bin_transformer_,
    ) = preprocess.preprocess_training_data(preprocess_params)
    assert ret_class_feat == CLASS_FEAT
    assert ret_pos_class == POS_CLASS
    assert ret_bin_transformer_ == None


def test_preprocess_training_data_X_np():
    preprocess_params = {
        "trainset": X_NP,
        "y": Y_NP,
        "class_feat": None,
        "pos_class": POS_CLASS,
        "feature_names": None,
        "n_discretize_bins": None,
        "verbosity": 0,
    }
    (
        ret_df,
        ret_class_feat,
        ret_pos_class,
        ret_bin_transformer_,
    ) = preprocess.preprocess_training_data(preprocess_params)
    x_ret_df = ret_df[ret_df.columns[:-1]]
    y_ret_df = ret_df[ret_df.columns[-1]]
    assert np.array_equal(x_ret_df.values, X_NP)
    assert np.array_equal(y_ret_df.values, Y_NP)
    assert ret_pos_class == POS_CLASS
    assert ret_bin_transformer_ == None


def test_preprocess_training_data_XY_np():
    preprocess_params = {
        "trainset": XY_NP,
        "y": None,
        "class_feat": NP_CLASS_FEAT,
        "pos_class": POS_CLASS,
        "feature_names": None,
        "n_discretize_bins": None,
        "verbosity": 0,
    }
    (
        ret_df,
        ret_class_feat,
        ret_pos_class,
        ret_bin_transformer_,
    ) = preprocess.preprocess_training_data(preprocess_params)
    assert np.array_equal(ret_df.values, XY_NP)
    assert ret_class_feat == NP_CLASS_FEAT
    assert ret_pos_class == POS_CLASS
    assert ret_bin_transformer_ == None


def test_preprocess_training_data_XY_rename_columns():

    # With xy
    preprocess_params = {
        "trainset": XY_NP,
        "y": None,
        "class_feat": CLASS_FEAT,
        "pos_class": POS_CLASS,
        "feature_names": DF.columns,
        "n_discretize_bins": None,
        "verbosity": 0,
    }
    (
        ret_df,
        ret_class_feat,
        ret_pos_class,
        ret_bin_transformer_,
    ) = preprocess.preprocess_training_data(preprocess_params)
    assert ret_df.equals(DF)
    assert ret_class_feat == CLASS_FEAT
    assert ret_pos_class == POS_CLASS
    assert ret_bin_transformer_ == None

    # With x_y
    preprocess_params = {
        "trainset": X_NP,
        "y": Y_NP,
        "class_feat": CLASS_FEAT,
        "pos_class": POS_CLASS,
        "feature_names": DF.drop(CLASS_FEAT, axis=1).columns,
        "n_discretize_bins": None,
        "verbosity": 0,
    }
    (
        ret_df,
        ret_class_feat,
        ret_pos_class,
        ret_bin_transformer_,
    ) = preprocess.preprocess_training_data(preprocess_params)
    x_ret_df = ret_df[ret_df.columns[:-1]]
    y_ret_df = ret_df[ret_df.columns[-1]]
    assert np.array_equal(x_ret_df.values, X_NP)
    assert np.array_equal(y_ret_df.values, Y_NP)
    assert x_ret_df.columns.tolist() == X_DF.columns.tolist()
    assert y_ret_df.name == CLASS_FEAT
    assert ret_class_feat == CLASS_FEAT
    assert ret_pos_class == POS_CLASS
    assert ret_bin_transformer_ == None

    # With x_y and don't rename class feature
    preprocess_params = {
        "trainset": X_NP,
        "y": Y_NP,
        "class_feat": None,
        "pos_class": POS_CLASS,
        "feature_names": DF.drop(CLASS_FEAT, axis=1).columns,
        "n_discretize_bins": None,
        "verbosity": 0,
    }
    (
        ret_df,
        ret_class_feat,
        ret_pos_class,
        ret_bin_transformer_,
    ) = preprocess.preprocess_training_data(preprocess_params)
    x_ret_df = ret_df[ret_df.columns[:-1]]
    y_ret_df = ret_df[ret_df.columns[-1]]
    assert np.array_equal(x_ret_df.values, X_NP)
    assert np.array_equal(y_ret_df.values, Y_NP)
    assert x_ret_df.columns.tolist() == X_DF.columns.tolist()
    assert y_ret_df.name == DEFAULT_CLASS_FEAT
    assert ret_class_feat == DEFAULT_CLASS_FEAT
    assert ret_pos_class == POS_CLASS
    assert ret_bin_transformer_ == None



from abc import ABC, abstractmethod
import numpy as np

from wittgenstein.check import _check_is_model_fit, _warn
import wittgenstein.base_functions as base_functions
import wittgenstein.preprocess as preprocess


class AbstractRulesetClassifier(ABC):
    def __init__(
        self,
        algorithm_name,
        prune_size=0.33,
        n_discretize_bins=10,
        max_rules=None,
        max_rule_conds=None,
        max_total_conds=None,
        random_state=None,
        verbosity=0,
    ):
        super().__init__()

        self.VALID_HYPERPARAMETERS = {
            "prune_size",
            "n_discretize_bins",
            "max_rules",
            "max_rule_conds",
            "max_total_conds",
            "random_state",
            "verbosity",
        }
        self.algorithm_name = algorithm_name
        self.prune_size = prune_size
        self.n_discretize_bins = n_discretize_bins
        self.max_rules = max_rules
        self.max_rule_conds = max_rule_conds
        self.max_total_conds = max_total_conds
        self.random_state = random_state
        self.verbosity = verbosity

        # This is to help keep sklearn ensemble happy should someone want use it
        self._estimator_type = "classifier"

    def __str__(self):
        """Return string representation."""
        params = str(self.get_params()) + ">"
        params = (
            params.replace(": ", "=")
            .replace("'", "")
            .replace("{", "(")
            .replace("}", ")")
        )
        return f"<{self.algorithm_name}{params}"

    __repr__ = __str__

    def out_model(self):
        """Print trained Ruleset model line-by-line: V represents 'or'; ^ represents 'and'."""
        if hasattr(self, "ruleset_"):
            self.ruleset_.out_pretty()
        else:
            print("no model fitted")

    def predict(self, X, give_reasons=False, feature_names=None):
        """Predict classes using a fit model.

        Parameters
        ----------
        X: DataFrame, numpy array, or other iterable
            Examples to make predictions on. All selected features of the model should be present.

        give_reasons : bool, default=False
            Whether to provide reasons for each prediction made.
        feature_names : list<str>, default=None
            Specify feature names for X to orient X's features with selected features.

        Returns
        -------
        list<bool>
            Predicted class labels for each row of X. True indicates positive predicted class, False negative class.

        Or, if give_reasons=True, returns

        tuple<list<bool>, <list<list<Rule>>>
            Tuple containing list of predictions and a list of the corresponding reasons for each prediction --
            for each positive prediction, a list of all the covering Rules, for negative predictions, an empty list.
        """

        _check_is_model_fit(self)

        # Preprocess prediction data
        preprocess_params = {
            "X": X,
            "class_feat": self.class_feat,
            "pos_class": self.pos_class,
            "bin_transformer_": self.bin_transformer_,
            "user_requested_feature_names": feature_names,
            "selected_features_": self.selected_features_,
            "trainset_features_": self.trainset_features_,
            "verbosity": self.verbosity,
        }

        X_df = preprocess.preprocess_prediction_data(preprocess_params)

        return self.ruleset_.predict(X_df, give_reasons=give_reasons)

    def score(self, X, y, score_function=base_functions.score_accuracy):

        """Score the performance of a fit model.

        X : DataFrame, numpy array, or other iterable
            Examples to score.
        y : Series, numpy array, or other iterable
            Class label actuals.

        score_function : function, default=score_accuracy
            Any scoring function that takes two parameters: actuals <iterable<bool>>,
            predictions <iterable<bool>>, where the elements represent class labels.
            This optional parameter is intended to be compatible with sklearn's scoring functions:
            https://scikit-learn.org/stable/modules/model_evaluation.html#classification-metrics
        """

        _check_is_model_fit(self)

        predictions = self.predict(X)
        actuals = [
            yi == self.pos_class for yi in preprocess._preprocess_y_score_data(y)
        ]
        return score_function(actuals, predictions)

    def predict_proba(self, X, give_reasons=False, feature_names=None):
        """Predict class probabilities of data using a fit model.

        Parameters
        ----------
            X: DataFrame, numpy array, or other iterable
                Examples to make predictions on. All selected features of the model should be present.

            give_reasons : bool, default=False
                Whether to provide reasons for each prediction made.
            feature_names : list<str>, default=None
                Specify feature names for X to orient X's features with selected features.

        Returns
        -------
        array
            Predicted class label probabilities for each row of X, ordered neg, pos.
            True indicates positive predicted class, False negative classes.

        or, if give_reasons=True:

        tuple< array, <list<list<Rule>> >
            Tuple containing array of predicted probabilities and a list of the corresponding reasons for each prediction--
            for each positive prediction, a list of all the covering Rules, for negative predictions, an empty list.
        """

        _check_is_model_fit(self)

        # Preprocess prediction data
        preprocess_params = {
            "X": X,
            "class_feat": self.class_feat,
            "pos_class": self.pos_class,
            "bin_transformer_": self.bin_transformer_,
            "user_requested_feature_names": feature_names,
            "selected_features_": self.selected_features_,
            "trainset_features_": self.trainset_features_,
            "verbosity": self.verbosity,
        }

        X_df = preprocess.preprocess_prediction_data(preprocess_params)

        # This is to help keep sklearn ensemble happy should someone want use it
        # self.classes_ = np.array([0, 1])
        self.classes_ = np.array([f"not {self.pos_class}", self.pos_class])

        return self.ruleset_.predict_proba(X_df, give_reasons=give_reasons)

    def recalibrate_proba(
        self,
        X_or_Xy,
        y=None,
        feature_names=None,
        min_samples=20,
        require_min_samples=True,
        discretize=True,
    ):
        """Recalibrate a classifier's probability estimations using unseen labeled data. May improve .predict_proba generalizability.
        Does not affect the underlying model or which predictions it makes -- only probability estimates.
        Use params min_samples and require_min_samples to select desired behavior.

        Note1: RunTimeWarning will occur as a reminder when min_samples and require_min_samples params might result in unintended effects.
        Note2: It is possible recalibrating could result in some positive .predict predictions with <0.5 .predict_proba positive probability.

        Xy: labeled data

        min_samples : int, default=20
            Required minimum number of samples per Rule.
            Use None to ignore minimum sampling requirement so long as at least one sample exists.
        require_min_samples : bool, default=True
            True: halt (with warning) in case min_samples not achieved for all Rules
            False: warn, but still replace Rules that have enough samples
        discretize : bool, default=True
            If the classifier has already fit bins, automatically discretize recalibrate_proba's training data
        """

        # Preprocess training data
        preprocess_params = {
            "X_or_Xy": X_or_Xy,
            "y": y,
            "class_feat": self.class_feat,
            "pos_class": self.pos_class,
            "bin_transformer_": self.bin_transformer_ if discretize else None,
            "user_requested_feature_names": feature_names,
            "min_samples": min_samples,
            "require_min_samples": require_min_samples,
            "verbosity": self.verbosity,
        }

        df = preprocess._preprocess_recalibrate_proba_data(preprocess_params)

        # Recalibrate
        base_functions.recalibrate_proba(
            self.ruleset_,
            Xy_df=df,
            class_feat=self.class_feat,
            pos_class=self.pos_class,
            min_samples=min_samples,
            require_min_samples=require_min_samples,
        )

    def get_params(self, deep=True):
        # parameter deep is a required artifact of sklearn compatability
        return {param: self.__dict__.get(param) for param in self.VALID_HYPERPARAMETERS}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self

    def _set_deprecated_fit_params(self, params):
        """Handle setting parameters passed to .fit that should have been passed to __init__"""
        found_deprecated_params = []
        for param, value in params.items():
            if param in self.VALID_HYPERPARAMETERS:
                found_deprecated_params.append(param)
                setattr(self, param, value)
        if found_deprecated_params:
            _warn(
                f"In the future, you should assign these parameters when initializating classifier instead of during model fitting: {found_deprecated_params}",
                DeprecationWarning,
                "irep/ripper",
                "fit",
            )



"""
Implementation of the RIPPERk algorithm for growing classification rulesets.
See https://www.let.rug.nl/nerbonne/teach/learning/cohen95fast.pdf
"""

# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

import copy
import math
import numpy as np

import pandas as pd

from wittgenstein import base, base_functions, preprocess
from .abstract_ruleset_classifier import AbstractRulesetClassifier
from .catnap import CatNap
from .check import _check_is_model_fit
from .base import Cond, Rule, Ruleset, rnd
from .base_functions import score_accuracy


class RIPPER(AbstractRulesetClassifier):
    """ Class for generating ruleset classification models.
        See Cohen (1995): https://www.let.rug.nl/nerbonne/teach/learning/cohen95fast.pdf
    """

    def __init__(
        self,
        k=2,
        dl_allowance=64,
        prune_size=0.33,
        n_discretize_bins=10,
        max_rules=None,
        max_rule_conds=None,
        max_total_conds=None,
        random_state=None,
        verbosity=0,
    ):
        """Create a RIPPER classifier.

        Parameters
        ----------
        k : int, default=2
            Number of RIPPERk optimization iterations.
        prune_size : float, default=.33
            Proportion of training set to be used for pruning.
        dl_allowance : int, default=64
            Terminate Ruleset grow phase early if a Ruleset description length is encountered
            that is more than this amount above the lowest description length so far encountered.
        n_discretize_bins : int, default=10
            Fit apparent numeric attributes into a maximum of n_discretize_bins discrete bins, inclusive on upper part of range. Pass None to disable auto-discretization.
        random_state : int, default=None
            Random seed for repeatable results.

        Limits for early-stopping. Intended for enhancing model interpretability and limiting training time on noisy datasets. Not specifically intended for use as a hyperparameter, since pruning already occurs during training, though it is certainly possible that tuning could improve model performance.
        max_rules : int, default=None
            Maximum number of rules.
        max_rule_conds : int, default=None
            Maximum number of conds per rule.
        max_total_conds : int, default=None
            Maximum number of total conds in entire ruleset.

        verbosity : int, default=0
            Output progress, model development, and/or computation. Each level includes the information belonging to lower-value levels.
               1: Show results of each major phase
               2: Show Ruleset grow/optimization steps
               3: Show Ruleset grow/optimization calculations
               4: Show Rule grow/prune steps
               5: Show Rule grow/prune calculations
        """

        AbstractRulesetClassifier.__init__(
            self,
            algorithm_name="RIPPER",
            prune_size=prune_size,
            n_discretize_bins=n_discretize_bins,
            max_rules=max_rules,
            max_rule_conds=max_rule_conds,
            max_total_conds=max_total_conds,
            random_state=random_state,
            verbosity=verbosity,
        )
        self.VALID_HYPERPARAMETERS.update({"k", "dl_allowance"})
        self.k = k
        self.dl_allowance = dl_allowance

    def __str__(self):
        """Return string representation of a RIPPER classifier."""
        params = str(self.get_params()) + ">"
        params = (
            params.replace(": ", "=")
            .replace("'", "")
            .replace("{", "(")
            .replace("}", ")")
        )
        return f"<RIPPER{params}"

    def out_model(self):
        """Print trained Ruleset model line-by-line: V represents 'or'; ^ represents 'and'."""
        super().out_model()

    def fit(
        self,
        trainset,
        y=None,
        class_feat=None,
        pos_class=None,
        feature_names=None,
        cn_optimize=True,
        **kwargs,
    ):
        """Fit a Ruleset model.

        Parameters
        ----------
        trainset : DataFrame, numpy array, or other iterable
            Training dataset. Optional whether to include or exclude class labels column.
        y : iterable of str, int, bool
            Class labels corresponding to trainset rows. Use if class labels aren't included in trainset.
        class_feat: str, int
            Column name or index of class feature. Use if class feature is still in trainset.
        pos_class : str, optional for boolean target, default=1 or True
            Name of positive class.
        feature_names : list<str>, optional, default=None
            Specify feature names. If None, feature names default to column names for a DataFrame, or indices in the case of indexed iterables such as an array or list.
        cn_optimize : bool, default=True
            Use algorithmic speed optimization.

        **kwargs
        --------
        The following parameters are moving to the RIPPER constructor (__init__) function. For the time-being, both the constructor and fit functions will accept them, but passing them here using .fit will be deprecated:

        k : int, default=2
            Number of RIPPERk optimization iterations.
        prune_size : float, default=.33
            Proportion of training set to be used for pruning.
        dl_allowance : int, default=64
            Terminate Ruleset grow phase early if a Ruleset description length is encountered
            that is more than this amount above the lowest description length so far encountered.
        n_discretize_bins : int, default=10
            Fit apparent numeric attributes into a maximum of n_discretize_bins discrete bins, inclusive on upper part of range. Pass None to disable auto-discretization.
        random_state : int, default=None
            Random seed for repeatable results.

        Limits for early-stopping. Intended for enhancing model interpretability and limiting training time on noisy datasets. Not specifically intended for use as a hyperparameter, since pruning already occurs during training, though it is certainly possible that tuning could improve model performance.
        max_rules : int, default=None
            Maximum number of rules.
        max_rule_conds : int, default=None
            Maximum number of conds per rule.
        max_total_conds : int, default=None
            Maximum number of total conds in entire ruleset.

        verbosity : int, default=0
            Output progress, model development, and/or computation. Each level includes the information belonging to lower-value levels.
               1: Show results of each major phase
               2: Show Ruleset grow/optimization steps
               3: Show Ruleset grow/optimization calculations
               4: Show Rule grow/prune steps
               5: Show Rule grow/prune calculations
        """

        ################
        # Stage 0: Setup
        ################

        # Handle any hyperparam deprecation
        self._set_deprecated_fit_params(kwargs)

        # Preprocess training data
        preprocess_params = {
            "trainset": trainset,
            "y": y,
            "class_feat": class_feat,
            "pos_class": pos_class,
            "feature_names": feature_names,
            "n_discretize_bins": self.n_discretize_bins,
            "verbosity": self.verbosity,
        }
        (
            df,
            self.class_feat,
            self.pos_class,
            self.bin_transformer_,
        ) = preprocess.preprocess_training_data(preprocess_params)

        # Create CatNap
        # possible minor speedup if pass cond_subset of only pos_class conds?
        if cn_optimize:
            self.cn = CatNap(
                df,
                feat_subset=None,
                cond_subset=None,
                class_feat=self.class_feat,
                pos_class=None,
            )

        # Split df into pos, neg classes
        pos_df, neg_df = base_functions.pos_neg_split(
            df, self.class_feat, self.pos_class
        )
        pos_df = pos_df.drop(self.class_feat, axis=1)
        neg_df = neg_df.drop(self.class_feat, axis=1)

        # Collect possible conds
        self._set_possible_conds(df)

        ###############################
        # Stage 1: Grow initial Ruleset
        ###############################

        self.ruleset_ = Ruleset()
        if cn_optimize:
            pos_idx = set(pos_df.index.tolist())
            neg_idx = set(neg_df.index.tolist())
            self.ruleset_ = self._grow_ruleset_cn(
                pos_idx,
                neg_idx,
                prune_size=self.prune_size,
                dl_allowance=self.dl_allowance,
                max_rules=self.max_rules,
                max_rule_conds=self.max_rule_conds,
                max_total_conds=self.max_total_conds,
                random_state=self.random_state,
            )
        else:
            self.ruleset_ = self._grow_ruleset(
                pos_df,
                neg_df,
                prune_size=self.prune_size,
                dl_allowance=self.dl_allowance,
                max_rules=self.max_rules,
                max_rule_conds=self.max_rule_conds,
                max_total_conds=self.max_total_conds,
                random_state=self.random_state,
            )
        if self.verbosity >= 1:
            print()
            print("GREW INITIAL RULESET:")
            self.ruleset_.out_pretty()
            print()

        ###########################
        # Stage 2: Optimize Ruleset
        ###########################

        for iter in range(1, self.k + 1):
            # Create new but reproducible random_state (if applicable)
            iter_random_state = (
                self.random_state + 100 if self.random_state is not None else None
            )
            # Run optimization iteration
            if self.verbosity >= 1:
                print(f"optimization run {iter} of {self.k}")
            if cn_optimize:
                newset = self._optimize_ruleset_cn(
                    self.ruleset_,
                    pos_idx,
                    neg_idx,
                    prune_size=self.prune_size,
                    random_state=iter_random_state,
                )
            else:
                newset = self._optimize_ruleset(
                    self.ruleset_,
                    pos_df,
                    neg_df,
                    prune_size=self.prune_size,
                    random_state=iter_random_state,
                )

            if self.verbosity >= 1:
                print()
                print("OPTIMIZED RULESET:")
                if self.verbosity >= 2:
                    print(
                        f"iteration {iter} of {self.k}\n modified rules {[i for i in range(len(self.ruleset_.rules)) if self.ruleset_.rules[i]!= newset.rules[i]]}"
                    )
                newset.out_pretty()
                print()

            if iter != self.k and self.ruleset_ == newset:
                if self.verbosity >= 1:
                    print("No changes were made. Halting optimization.")
                break
            else:
                self.ruleset_ = newset

        #############################################
        # Stage 3: Cover any last remaining positives
        #############################################

        if cn_optimize:
            self._cover_remaining_positives_cn(
                df,
                max_rules=self.max_rules,
                max_rule_conds=self.max_rule_conds,
                max_total_conds=self.max_total_conds,
                random_state=self.random_state,
            )
        else:
            self._cover_remaining_positives(
                df,
                max_rules=self.max_rules,
                max_rule_conds=self.max_rule_conds,
                max_total_conds=self.max_total_conds,
                random_state=self.random_state,
            )

        #################################################
        # Stage 4: Remove any rules that don't improve dl
        #################################################

        if self.verbosity >= 2:
            print("Optimizing dl...")
        if cn_optimize:
            mdl_subset, _ = _rs_total_bits_cn(
                self.cn,
                self.ruleset_,
                self.ruleset_.possible_conds,
                pos_idx,
                neg_idx,
                bestsubset_dl=True,
                ret_bestsubset=True,
                verbosity=self.verbosity,
            )
        else:
            mdl_subset, _ = _rs_total_bits(
                self.ruleset_,
                self.ruleset_.possible_conds,
                pos_df,
                neg_df,
                bestsubset_dl=True,
                ret_bestsubset=True,
                verbosity=self.verbosity,
            )
        self.ruleset_ = mdl_subset
        if self.verbosity >= 1:
            print("FINAL RULESET:")
            self.ruleset_.out_pretty()
            print()

        # Issue warning if Ruleset is universal or empty
        self.ruleset_._check_allpos_allneg(warn=True, warnstack=[("ripper", "fit")])

        # Set Ruleset features
        self.selected_features_ = self.ruleset_.get_selected_features()
        self.trainset_features_ = df.drop(self.class_feat, axis=1).columns.tolist()

        # Fit probas
        self.recalibrate_proba(
            df, min_samples=None, require_min_samples=False, discretize=False
        )

        # Cleanup
        if cn_optimize:
            del self.cn

    def score(self, X, y, score_function=score_accuracy):
        """Score the performance of a fit model.

        X : DataFrame, numpy array, or other iterable
            Examples to score.
        y : Series, numpy array, or other iterable
            Class label actuals.

        score_function : function, default=score_accuracy
            Any scoring function that takes two parameters: actuals <iterable<bool>>, predictions <iterable<bool>>, where the elements represent class labels.
            this optional parameter is intended to be compatible with sklearn's scoring functions: https://scikit-learn.org/stable/modules/model_evaluation.html#classification-metrics
        """

        _check_is_model_fit(self)

        predictions = self.predict(X)
        actuals = [yi == self.pos_class for yi in base_functions.aslist(y)]
        return score_function(actuals, predictions)

    def _set_theory_dl_lookup(self, df, size=15, verbosity=0):
        """Precalculate rule theory dls for various-sized rules."""

        self.dl_dict = {}

        temp = Ruleset()
        temp._set_possible_conds(df, df)
        possible_conds = temp.possible_conds

        for n in range(1, size + 1):
            rule = Rule([Cond("_", "_")] * n)
            dl = _r_theory_bits(
                rule, possible_conds, bits_dict=None, verbosity=verbosity
            )
            self.dl_dict[n] = dl
            if verbosity >= 2:
                print(f"updated dl for rule size {n}: {dl}")

    def _grow_ruleset(
        self,
        pos_df,
        neg_df,
        prune_size,
        dl_allowance,
        initial_ruleset=None,
        max_rules=None,
        max_rule_conds=None,
        max_total_conds=None,
        random_state=None,
    ):
        """Grow a Ruleset with pruning."""
        pos_remaining = pos_df.copy()
        neg_remaining = neg_df.copy()

        if initial_ruleset is None:
            ruleset = Ruleset()
            ruleset._set_possible_conds(pos_df, neg_df)
        else:
            ruleset = copy.deepcopy(initial_ruleset)

        ruleset_dl = None
        mdl = None  # Minimum encountered description length (in bits)
        dl_diff = 0
        if self.verbosity >= 2:
            print("growing ruleset...")
            print()
        while len(pos_remaining) > 0 and dl_diff <= self.dl_allowance:

            # If applicable, check for user-specified early stopping
            if stop_early(ruleset, max_rules, max_total_conds):
                break

            # Grow-prune split remaining uncovered examples
            pos_growset, pos_pruneset = base_functions.df_shuffled_split(
                pos_remaining, (1 - prune_size), random_state=random_state
            )
            neg_growset, neg_pruneset = base_functions.df_shuffled_split(
                neg_remaining, (1 - prune_size), random_state=random_state
            )
            if self.verbosity >= 2:
                print(
                    f"pos_growset {len(pos_growset)} pos_pruneset {len(pos_pruneset)}"
                )
                print(
                    f"neg_growset {len(neg_growset)} neg_pruneset {len(neg_pruneset)}"
                )
            if len(pos_growset) == 0:
                break  # Probably safe, but a little dicey to only check pos_growset.

            # Grow Rule
            grown_rule = base_functions.grow_rule(
                pos_growset,
                neg_growset,
                ruleset.possible_conds,
                max_rule_conds=max_rule_conds,
                verbosity=self.verbosity,
            )
            if grown_rule.isempty():
                break  # Generated an empty rule b/c no good conds exist

            # Prune Rule
            pruned_rule = base_functions.prune_rule(
                grown_rule,
                _RIPPER_growphase_prune_metric,
                pos_pruneset,
                neg_pruneset,
                verbosity=self.verbosity,
            )

            # Add rule; calculate new description length
            ruleset.add(
                pruned_rule
            )  # Unlike IREP, IREP*/RIPPER stopping condition is inclusive: "After each rule is added, the total description length of the rule set and examples is computed."
            if self.verbosity >= 2:
                print(f"updated ruleset: {ruleset.truncstr(direction='right')}")
                print()

            if ruleset_dl is None:  # First Rule to be added
                rule_dl = _r_theory_bits(
                    pruned_rule, ruleset.possible_conds, verbosity=self.verbosity
                )
                theory_dl = rule_dl
                data_dl = _exceptions_bits(
                    ruleset, pos_df, neg_df, verbosity=self.verbosity
                )
                ruleset_dl = theory_dl + data_dl
                mdl = ruleset_dl
            else:
                rule_dl = _r_theory_bits(
                    pruned_rule, ruleset.possible_conds, verbosity=self.verbosity
                )
                theory_dl += rule_dl
                data_dl = _exceptions_bits(
                    ruleset, pos_df, neg_df, verbosity=self.verbosity
                )
                ruleset_dl = theory_dl + data_dl
                dl_diff = ruleset_dl - mdl

            if self.verbosity >= 3:
                print(f"rule dl: {rnd(rule_dl)}")
                print(f"updated theory dl: {rnd(theory_dl)}")
                print(f"exceptions: {rnd(data_dl)}")
                print(f"total dl: {rnd(ruleset_dl)}")
                if dl_diff <= self.dl_allowance:
                    print(
                        f"mdl {rnd(mdl)} (diff {rnd(dl_diff)} <= {rnd(self.dl_allowance)})"
                    )
                else:
                    print(
                        f"mdl {rnd(mdl)} dl-halt: diff {rnd(dl_diff)} exceeds allowance ({rnd(self.dl_allowance)})"
                    )

            mdl = ruleset_dl if ruleset_dl < mdl else mdl

            # Remove covered examples
            pos_remaining, neg_remaining = base_functions.rm_covered(
                pruned_rule, pos_remaining, neg_remaining
            )

            if self.verbosity >= 3:
                print(
                    f"examples remaining: {len(pos_remaining)} pos, {len(neg_remaining)} neg"
                )
                print()

        # If applicable, trim total conds
        ruleset.trim_conds(max_total_conds=max_total_conds)

        return ruleset

    def _grow_ruleset_cn(
        self,
        pos_idx,
        neg_idx,
        prune_size,
        dl_allowance,
        initial_ruleset=None,
        max_rules=None,
        max_rule_conds=None,
        max_total_conds=None,
        random_state=None,
    ):
        """Grow a Ruleset with pruning."""

        # Initialize new Ruleset
        if initial_ruleset is None:
            ruleset = Ruleset()
            ruleset.possible_conds = self.possible_conds  # cn.conds
        else:
            ruleset = copy.deepcopy(initial_ruleset)

        pos_remaining_idx = pos_idx
        neg_remaining_idx = neg_idx
        ruleset_dl = None
        mdl = None  # Minimum encountered description length (in bits)
        dl_diff = 0
        if self.verbosity >= 2:
            print("growing ruleset...")
            print()
        while len(pos_remaining_idx) > 0 and dl_diff <= self.dl_allowance:

            # If applicable, check for user-specified early stopping
            if (max_rules is not None and len(ruleset.rules) >= max_rules) or (
                max_total_conds is not None and ruleset.count_conds() >= max_total_conds
            ):
                break

            # Grow-prune split remaining uncovered examples
            pos_growset_idx, pos_pruneset_idx = base_functions.random_split(
                pos_remaining_idx,
                (1 - prune_size),
                res_type=set,
                random_state=random_state,
            )
            neg_growset_idx, neg_pruneset_idx = base_functions.random_split(
                neg_remaining_idx,
                (1 - prune_size),
                res_type=set,
                random_state=random_state,
            )
            if self.verbosity >= 2:
                print(
                    f"pos_growset {len(pos_growset_idx)} pos_pruneset {len(pos_pruneset_idx)}"
                )
                print(
                    f"neg_growset {len(neg_growset_idx)} neg_pruneset {len(neg_pruneset_idx)}"
                )
            if len(pos_growset_idx) == 0:
                break  # Probably safe, but a little dicey to only check pos_growset.

            # Grow Rule
            grown_rule = base_functions.grow_rule_cn(
                self.cn,
                pos_growset_idx,
                neg_growset_idx,
                initial_rule=Rule(),
                max_rule_conds=max_rule_conds,
                verbosity=self.verbosity,
            )
            if grown_rule.isempty():
                break  # Generated an empty rule b/c no good conds exist

            # Prune Rule
            pruned_rule = base_functions.prune_rule_cn(
                self.cn,
                grown_rule,
                _RIPPER_growphase_prune_metric_cn,
                pos_pruneset_idx,
                neg_pruneset_idx,
                verbosity=self.verbosity,
            )

            # Add rule; calculate new description length
            ruleset.add(
                pruned_rule
            )  # Unlike IREP, IREP*/RIPPER stopping condition is inclusive: "After each rule is added, the total description length of the rule set and examples is computed."
            if self.verbosity >= 2:
                print(f"updated ruleset: {ruleset.truncstr(direction='right')}")
                print()

            if ruleset_dl is None:  # First Rule to be added
                rule_dl = _r_theory_bits(
                    pruned_rule, ruleset.possible_conds, verbosity=self.verbosity
                )
                theory_dl = rule_dl
                data_dl = _exceptions_bits_cn(
                    self.cn, ruleset, pos_idx, neg_idx, verbosity=self.verbosity
                )
                ruleset_dl = theory_dl + data_dl
                mdl = ruleset_dl
            else:
                rule_dl = _r_theory_bits(
                    pruned_rule, ruleset.possible_conds, verbosity=self.verbosity
                )
                theory_dl += rule_dl
                data_dl = _exceptions_bits_cn(
                    self.cn, ruleset, pos_idx, neg_idx, verbosity=self.verbosity
                )
                ruleset_dl = theory_dl + data_dl
                dl_diff = ruleset_dl - mdl

            if self.verbosity >= 3:
                print(f"rule dl: {rnd(rule_dl)}")
                print(f"updated theory dl: {rnd(theory_dl)}")
                print(f"exceptions: {rnd(data_dl)}")
                print(f"total dl: {rnd(ruleset_dl)}")
                if dl_diff <= self.dl_allowance:
                    print(
                        f"mdl {rnd(mdl)} (diff {rnd(dl_diff)} <= {rnd(self.dl_allowance)})"
                    )
                else:
                    print(
                        f"mdl {rnd(mdl)} dl-halt: diff {rnd(dl_diff)} exceeds allowance ({rnd(self.dl_allowance)})"
                    )

            mdl = ruleset_dl if ruleset_dl < mdl else mdl

            # Remove covered examples
            pos_remaining_idx, neg_remaining_idx = base_functions.rm_rule_covers_cn(
                self.cn, pruned_rule, pos_remaining_idx, neg_remaining_idx
            )

            if self.verbosity >= 3:
                print(
                    f"examples remaining: {len(pos_remaining_idx)} pos, {len(neg_remaining_idx)} neg"
                )
                print()

        # If applicable, trim total conds
        ruleset.trim_conds(max_total_conds=max_total_conds)

        return ruleset

    def _optimize_ruleset(
        self,
        ruleset,
        pos_df,
        neg_df,
        prune_size,
        max_rule_conds=None,
        random_state=None,
    ):
        """Optimization phase."""

        if self.verbosity >= 2:
            print("optimizing ruleset...")
            print()

        pos_remaining = pos_df.copy()
        neg_remaining = neg_df.copy()
        original_ruleset = copy.deepcopy(ruleset)
        if self.verbosity >= 4:
            print("calculate original ruleset potential dl...")
        original_dl = _rs_total_bits(
            original_ruleset,
            original_ruleset.possible_conds,
            pos_df,
            neg_df,
            bestsubset_dl=True,
            verbosity=self.verbosity,
        )
        if self.verbosity >= 3:
            print(f"original ruleset potential dl: {rnd(original_dl)}")
            print()
        new_ruleset = copy.deepcopy(ruleset)

        for i, rule in enumerate(original_ruleset.rules):
            pos_growset, pos_pruneset = base_functions.df_shuffled_split(
                pos_remaining, (1 - prune_size), random_state=random_state
            )
            neg_growset, neg_pruneset = base_functions.df_shuffled_split(
                neg_remaining, (1 - prune_size), random_state=random_state
            )
            if len(pos_growset) == 0:
                break  # Possible where optimization run > 1

            # Create alternative rules
            if self.verbosity >= 4:
                print(
                    f"creating replacement for {i} of {len(original_ruleset.rules)}: {ruleset.rules[i]}"
                )
            g_replacement = base_functions.grow_rule(
                pos_growset,
                neg_growset,
                original_ruleset.possible_conds,
                initial_rule=Rule(),
                max_rule_conds=max_rule_conds,
                verbosity=self.verbosity,
            )
            replacement_ruleset = Ruleset(
                base_functions.i_replaced(original_ruleset.rules, i, g_replacement)
            )
            pr_replacement = base_functions.prune_rule(
                g_replacement,
                _RIPPER_optimization_prune_metric,
                pos_pruneset,
                neg_pruneset,
                eval_index_on_ruleset=(i, replacement_ruleset),
                verbosity=self.verbosity,
            )
            replacement_ruleset = Ruleset(
                base_functions.i_replaced(original_ruleset.rules, i, pr_replacement)
            )
            if self.verbosity >= 3:
                print(f"grew replacement {g_replacement}")
                print(f"pruned replacement is {pr_replacement}")

            if self.verbosity >= 3:
                print(
                    f"creating revision for {i} of {len(original_ruleset.rules)}: {ruleset.rules[i]}"
                )
            g_revision = base_functions.grow_rule(
                pos_growset,
                neg_growset,
                original_ruleset.possible_conds,
                initial_rule=ruleset.rules[i],
                max_rule_conds=max_rule_conds,
                verbosity=self.verbosity,
            )
            revision_ruleset = Ruleset(
                base_functions.i_replaced(original_ruleset.rules, i, g_revision)
            )
            pr_revision = base_functions.prune_rule(
                g_revision,
                _RIPPER_optimization_prune_metric,
                pos_pruneset,
                neg_pruneset,
                eval_index_on_ruleset=(i, revision_ruleset),
                verbosity=self.verbosity,
            )
            revision_ruleset = Ruleset(
                base_functions.i_replaced(original_ruleset.rules, i, pr_revision)
            )
            if self.verbosity >= 3:
                print(f"grew revision {g_replacement}")
                print(f"pruned revision is {pr_replacement}")
                print()

            # Calculate alternative Rulesets' respective lowest potential dls to identify the best version
            if self.verbosity >= 3:
                print(
                    f"calculate potential dl for ds with replacement {pr_replacement}"
                )
            replacement_dl = (
                _rs_total_bits(
                    replacement_ruleset,
                    original_ruleset.possible_conds,
                    pos_df,
                    neg_df,
                    bestsubset_dl=True,
                    verbosity=self.verbosity,
                )
                if pr_replacement != rule
                else original_dl
            )
            if self.verbosity >= 3:
                print(f"calculate potential dl for ds with revision {pr_revision}")
            revision_dl = (
                _rs_total_bits(
                    revision_ruleset,
                    original_ruleset.possible_conds,
                    pos_df,
                    neg_df,
                    bestsubset_dl=True,
                    verbosity=self.verbosity,
                )
                if pr_revision != rule
                else original_dl
            )
            best_rule = [rule, pr_replacement, pr_revision][
                base_functions.argmin([original_dl, replacement_dl, revision_dl])
            ]

            if self.verbosity >= 2:
                print(f"\nrule {i+1} of {len(original_ruleset.rules)}")
                rep_str = (
                    pr_replacement.__str__() if pr_replacement != rule else "unchanged"
                )
                rev_str = pr_revision.__str__() if pr_revision != rule else "unchanged"
                best_str = best_rule.__str__() if best_rule != rule else "unchanged"
                if self.verbosity == 2:
                    print(f"original: {rule}")
                    print(f"replacement: {rep_str}")
                    print(f"revision: {rev_str}")
                    print(f"*best: {best_str}")
                    if best_rule in new_ruleset:
                        print(
                            f"best already included in optimization -- retaining original"
                        )
                    print()
                else:
                    print(f"original: {rule}) | {rnd(original_dl)} bits")
                    print(f"replacement: {rep_str} | {rnd(replacement_dl)} bits")
                    print(f"revision: {rev_str} | {rnd(revision_dl)} bits")
                    print(
                        f"*best: {best_str} | {rnd(min([replacement_dl, revision_dl, original_dl]))} bits"
                    )
                    if best_rule in new_ruleset:
                        print(
                            f"best already included in optimization -- retaining original"
                        )
                    print()
            if best_rule not in new_ruleset:
                new_ruleset.rules[i] = best_rule
            else:
                new_ruleset.rules[i] = rule

            # Remove covered examples
            pos_remaining, neg_remaining = base_functions.rm_covered(
                rule, pos_remaining, neg_remaining
            )
            if self.verbosity >= 3:
                print(
                    f"examples remaining: {len(pos_remaining)} pos, {len(neg_remaining)} neg"
                )
                print()

            # If there are no pos data remaining to train optimization (could happen if optimization run >1), keep remaining rules the same
            if len(pos_remaining) == 0:
                break

        return new_ruleset

    def _optimize_ruleset_cn(
        self,
        ruleset,
        pos_idx,
        neg_idx,
        prune_size,
        max_rule_conds=None,
        random_state=None,
    ):
        """Optimization phase."""

        if self.verbosity >= 2:
            print("optimizing ruleset...")
            print()

        pos_remaining_idx = pos_idx
        neg_remaining_idx = neg_idx
        original_ruleset = copy.deepcopy(ruleset)
        if self.verbosity >= 4:
            print("calculate original ruleset potential dl...")
        original_dl = _rs_total_bits_cn(
            self.cn,
            original_ruleset,
            original_ruleset.possible_conds,
            pos_idx,
            neg_idx,
            bestsubset_dl=True,
            verbosity=self.verbosity,
        )
        if self.verbosity >= 3:
            print(f"original ruleset potential dl: {rnd(original_dl)}")
            print()
        new_ruleset = copy.deepcopy(ruleset)

        for i, rule in enumerate(original_ruleset.rules):
            pos_growset_idx, pos_pruneset_idx = base_functions.set_shuffled_split(
                pos_remaining_idx, (1 - prune_size), random_state=random_state
            )
            neg_growset_idx, neg_pruneset_idx = base_functions.set_shuffled_split(
                neg_remaining_idx, (1 - prune_size), random_state=random_state
            )
            if len(pos_growset_idx) == 0:
                break  # Possible where optimization run > 1

            # Create alternative rules
            if self.verbosity >= 4:
                print(
                    f"creating replacement for {i} of {len(original_ruleset.rules)}: {ruleset.rules[i]}"
                )
            # g_replacement = base_functions.grow_rule(pos_growset, neg_growset, original_ruleset.possible_conds, initial_rule=Rule(), max_rule_conds=max_rule_conds, verbosity=self.verbosity)
            g_replacement = base_functions.grow_rule_cn(
                self.cn,
                pos_growset_idx,
                neg_growset_idx,
                initial_rule=Rule(),
                max_rule_conds=max_rule_conds,
                verbosity=self.verbosity,
            )
            replacement_ruleset = Ruleset(
                base_functions.i_replaced(original_ruleset.rules, i, g_replacement)
            )
            # pr_replacement = base_functions.prune_rule(g_replacement, _RIPPER_optimization_prune_metric, pos_pruneset, neg_pruneset, eval_index_on_ruleset=(i,replacement_ruleset), verbosity=self.verbosity)
            pr_replacement = base_functions.prune_rule_cn(
                self.cn,
                g_replacement,
                _RIPPER_optimization_prune_metric_cn,
                pos_pruneset_idx,
                neg_pruneset_idx,
                eval_index_on_ruleset=(i, replacement_ruleset),
                verbosity=self.verbosity,
            )
            replacement_ruleset = Ruleset(
                base_functions.i_replaced(original_ruleset.rules, i, pr_replacement)
            )
            if self.verbosity >= 3:
                print(f"grew replacement {g_replacement}")
                print(f"pruned replacement is {pr_replacement}")

            if self.verbosity >= 3:
                print(
                    f"creating revision for {i} of {len(original_ruleset.rules)}: {ruleset.rules[i]}"
                )
            # g_revision = base_functions.grow_rule(pos_growset, neg_growset, original_ruleset.possible_conds, initial_rule=ruleset.rules[i], max_rule_conds=max_rule_conds, verbosity=self.verbosity)
            g_revision = base_functions.grow_rule_cn(
                self.cn,
                pos_growset_idx,
                neg_growset_idx,
                initial_rule=ruleset.rules[i],
                max_rule_conds=max_rule_conds,
                verbosity=self.verbosity,
            )
            revision_ruleset = Ruleset(
                base_functions.i_replaced(original_ruleset.rules, i, g_revision)
            )
            # pr_revision = base_functions.prune_rule(g_revision, _RIPPER_optimization_prune_metric, pos_pruneset, neg_pruneset, eval_index_on_ruleset=(i,revision_ruleset), verbosity=self.verbosity)
            pr_revision = base_functions.prune_rule_cn(
                self.cn,
                g_revision,
                _RIPPER_optimization_prune_metric_cn,
                pos_pruneset_idx,
                neg_pruneset_idx,
                eval_index_on_ruleset=(i, revision_ruleset),
                verbosity=self.verbosity,
            )
            revision_ruleset = Ruleset(
                base_functions.i_replaced(original_ruleset.rules, i, pr_revision)
            )
            if self.verbosity >= 3:
                print(f"grew revision {g_replacement}")
                print(f"pruned revision is {pr_replacement}")
                print()

            # Calculate alternative Rulesets' respective lowest potential dls to identify the best version
            if self.verbosity >= 3:
                print(
                    f"calculate potential dl for ds with replacement {pr_replacement}"
                )
            replacement_dl = (
                _rs_total_bits_cn(
                    self.cn,
                    replacement_ruleset,
                    original_ruleset.possible_conds,
                    pos_idx,
                    neg_idx,
                    bestsubset_dl=False,
                    ret_bestsubset=False,
                    verbosity=0,
                )
                if pr_replacement != rule
                else original_dl
            )
            if self.verbosity >= 3:
                print(f"calculate potential dl for ds with revision {pr_revision}")
            revision_dl = (
                _rs_total_bits_cn(
                    self.cn,
                    revision_ruleset,
                    original_ruleset.possible_conds,
                    pos_idx,
                    neg_idx,
                    bestsubset_dl=False,
                    ret_bestsubset=False,
                    verbosity=0,
                )
                if pr_revision != rule
                else original_dl
            )
            best_rule = [rule, pr_replacement, pr_revision][
                base_functions.argmin([original_dl, replacement_dl, revision_dl])
            ]

            if self.verbosity >= 2:
                print(f"\nrule {i+1} of {len(original_ruleset.rules)}")
                rep_str = (
                    pr_replacement.__str__() if pr_replacement != rule else "unchanged"
                )
                rev_str = pr_revision.__str__() if pr_revision != rule else "unchanged"
                best_str = best_rule.__str__() if best_rule != rule else "unchanged"
                if self.verbosity == 2:
                    print(f"original: {rule}")
                    print(f"replacement: {rep_str}")
                    print(f"revision: {rev_str}")
                    print(f"*best: {best_str}")
                    if best_rule in new_ruleset:
                        print(
                            f"best already included in optimization -- retaining original"
                        )
                    print()
                else:
                    print(f"original: {rule}) | {rnd(original_dl)} bits")
                    print(f"replacement: {rep_str} | {rnd(replacement_dl)} bits")
                    print(f"revision: {rev_str} | {rnd(revision_dl)} bits")
                    print(
                        f"*best: {best_str} | {rnd(min([replacement_dl, revision_dl, original_dl]))} bits"
                    )
                    if best_rule in new_ruleset:
                        print(
                            f"best already included in optimization -- retaining original"
                        )
                    print()
            if best_rule not in new_ruleset:
                new_ruleset.rules[i] = best_rule
            else:
                new_ruleset.rules[i] = rule

            # Remove covered examples
            pos_remaining_idx, neg_remaining_idx = base_functions.rm_rule_covers_cn(
                self.cn, rule, pos_remaining_idx, neg_remaining_idx
            )
            if self.verbosity >= 3:
                print(
                    f"examples remaining: {len(pos_remaining_idx)} pos, {len(neg_remaining_idx)} neg"
                )
                print()

            # If there are no pos data remaining to train optimization (could happen if optimization run >1), keep remaining rules the same
            if len(pos_remaining_idx) == 0:
                break

        return new_ruleset

    def _set_possible_conds(self, df):
        self.possible_conds = []
        for feat in df.columns.values:
            if feat != self.class_feat:
                for val in df[feat].unique():
                    self.possible_conds.append(Cond(feat, val))

    def _cover_remaining_positives(
        self,
        df,
        max_rules=None,
        max_rule_conds=None,
        max_total_conds=None,
        random_state=None,
    ):
        """Stage 3: Post-optimization, cover any remaining uncovered positives."""
        pos_remaining, neg_remaining = base_functions.pos_neg_split(
            df, self.class_feat, self.pos_class
        )
        pos_remaining = pos_remaining.drop(self.class_feat, axis=1)
        neg_remaining = neg_remaining.drop(self.class_feat, axis=1)
        pos_remaining, neg_remaining = base_functions.rm_covered(
            self.ruleset_, pos_remaining, neg_remaining
        )
        if len(pos_remaining) >= 1:
            if self.verbosity >= 2:
                print(f"{len(pos_remaining)} pos left. Growing final rules...")
            newset = self._grow_ruleset(
                pos_remaining,
                neg_remaining,
                initial_ruleset=self.ruleset_,
                prune_size=self.prune_size,
                dl_allowance=self.dl_allowance,
                max_rules=max_rules,
                max_rule_conds=max_rule_conds,
                max_total_conds=max_total_conds,
                random_state=random_state,
            )
            if self.verbosity >= 1:
                print("GREW FINAL RULES")
                newset.out_pretty()
                print()
            self.ruleset_ = newset
        else:
            if self.verbosity >= 1:
                print("All pos covered\n")

    def _cover_remaining_positives_cn(
        self,
        df,
        max_rules=None,
        max_rule_conds=None,
        max_total_conds=None,
        random_state=None,
    ):
        """Stage 3: Post-optimization, cover any remaining uncovered positives."""
        pos_remaining_idx, neg_remaining_idx = self.cn.pos_idx_neg_idx(
            df, self.class_feat, self.pos_class
        )

        if len(pos_remaining_idx) >= 1:
            if self.verbosity >= 2:
                print(f"{len(pos_remaining_idx)} pos left. Growing final rules...")
            newset = self._grow_ruleset_cn(
                pos_remaining_idx,
                neg_remaining_idx,
                initial_ruleset=self.ruleset_,
                prune_size=self.prune_size,
                dl_allowance=self.dl_allowance,
                max_rules=max_rules,
                max_rule_conds=max_rule_conds,
                max_total_conds=max_total_conds,
                random_state=random_state,
            )
            if self.verbosity >= 1:
                print("GREW FINAL RULES")
                newset.out_pretty()
                print()
            self.ruleset_ = newset
        else:
            if self.verbosity >= 1:
                print("All positives covered\n")

    def get_params(self, deep=True):
        return {param: self.__dict__.get(param) for param in self.VALID_HYPERPARAMETERS}

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            # if parameter in self.VALID_HYPERPARAMETERS:
            setattr(self, parameter, value)
        return self


###################################
##### RIPPER-specific Metrics #####
###################################


def _RIPPER_growphase_prune_metric(rule, pos_pruneset, neg_pruneset):
    """RIPPER/IREP* prune metric.
    Returns the prune value of a candidate Rule.

    Cohen's formula is (p-n) / (p+n).
    Unclear from the paper how they handle divzero (where p+n=0), so I Laplaced it.
    Weka's solution was to modify the formula to (p+1)/(p+n+2), but running with this because the (non-NaN) values I got appeared closer to those of the original formula.
    """
    # I imagine Weka's is 1/2 because that's closer to a 50-50 class distribution?
    p = rule.num_covered(pos_pruneset)
    n = rule.num_covered(neg_pruneset)
    return (p - n + 1) / (p + n + 1)


def _RIPPER_growphase_prune_metric_cn(cn, rule, pos_pruneset_idx, neg_pruneset_idx):
    """RIPPER/IREP* prune metric.
    Returns the prune value of a candidate Rule.

    Cohen's formula is (p-n) / (p+n).
    Unclear from the paper how they handle divzero (where p+n=0), so I Laplaced it.
    Weka's solution was to modify the formula to (p+1)/(p+n+2), but the (non-NaN) values I got appeared closer to those of the original formula.
    """
    # I imagine Weka's is 1/2 because that's closer to a 50-50 class distribution?
    p = len(cn.rule_covers(rule, pos_pruneset_idx))
    n = len(cn.rule_covers(rule, neg_pruneset_idx))
    return (p - n + 1) / (p + n + 1)


def _RIPPER_optimization_prune_metric(rule, pos_pruneset, neg_pruneset):
    return base_functions._accuracy(rule, pos_pruneset, neg_pruneset)


def _RIPPER_optimization_prune_metric_cn(cn, rule, pos_pruneset_idx, neg_pruneset_idx):
    return base_functions._rule_accuracy_cn(
        cn, rule, pos_pruneset_idx, neg_pruneset_idx
    )


def _r_theory_bits(rule, possible_conds, bits_dict=None, verbosity=0):
    """Returns description length (in bits) for a single Rule."""

    if hasattr(rule, "dl"):
        return rule.dl
    else:
        # if type(rule) != Rule:
        #    raise TypeError(f'param rule in _r_theory_bits is type {type(rule)}; it should be type Rule')
        k = len(rule.conds)  # Number of rule conditions
        n = len(possible_conds)  # Number of possible conditions
        pr = k / n

        S = k * math.log2(1 / pr) + (n - k) * math.log2((1 / (1 - pr)))  # S(n, k, pr)
        K = math.log2(k)  # Number bits need to send integer k
        rule_dl = 0.5 * (
            K + S
        )  # Divide by 2 a la Quinlan. Cohen: "to adjust for possible redundency in attributes"
        if verbosity >= 5:
            print(
                f"rule theory bits| {rule} k {k} n {n} pr {rnd(pr)}: {rnd(rule_dl)} bits"
            )

        # rule.dl = rule_dl
        return rule_dl


def _rs_theory_bits(ruleset, possible_conds, verbosity=0):
    """Returns theory description length (in bits) for a Ruleset."""

    # if type(ruleset) != Ruleset:
    #    raise TypeError(f'param ruleset in _rs_theory_bits should be type Ruleset')
    total = 0
    for rule in ruleset.rules:
        total += _r_theory_bits(rule, possible_conds, verbosity=verbosity)
        # total += rule_bits(rule, possible_conds, rem_pos, rem_neg, verbosity=verbosity)
        # rem_pos, rem_neg = base.rm_covered(rule, rem_pos, rem_neg)
    if verbosity >= 5:
        print(f"ruleset theory bits| {rnd(total)}")

    # ruleset.dl = total
    return total


def _exceptions_bits(ruleset, pos_df, neg_df, verbosity=0):
    """Returns description length (in bits) for exceptions to a Ruleset's coverage."""

    if type(ruleset) != Ruleset:
        raise TypeError(
            f"to avoid double-counting, _exceptions_bits should calculate exceptions over entire set of rules with type Ruleset"
        )
    N = len(pos_df) + len(neg_df)  # Total number of examples
    p = ruleset.num_covered(pos_df) + ruleset.num_covered(
        neg_df
    )  # Total number of examples classified as positive = total covered
    fp = ruleset.num_covered(
        neg_df
    )  # Number false positives = negatives covered by the ruleset
    fn = len(pos_df) - ruleset.num_covered(
        pos_df
    )  # Number false negatives = positives not covered by the ruleset
    exceptions_dl = math.log2(base_functions.nCr(p, fp)) + math.log2(
        base_functions.nCr((N - p), fn)
    )
    if verbosity >= 5:
        print(
            f"exceptions_bits| {ruleset.truncstr()}: \n N {N} p {p} fp {fp} fn {fn}: exceptions_bits {rnd(exceptions_dl)}"
        )

    return exceptions_dl


def _exceptions_bits_cn(cn, ruleset, pos_idx, neg_idx, verbosity=0):
    """Returns description length (in bits) for exceptions to a Ruleset's coverage."""

    # if type(ruleset) != Ruleset:
    #    raise TypeError(f'to avoid double-counting, _exceptions_bits should calculate exceptions over entire set of rules with type Ruleset')
    N = len(pos_idx) + len(neg_idx)  # Total number of examples
    pos_cov = cn.ruleset_covers(ruleset, subset=pos_idx)
    neg_cov = cn.ruleset_covers(ruleset, subset=neg_idx)
    p = len(pos_cov) + len(
        neg_cov
    )  # Total number of examples classified as positive = total covered
    fp = len(neg_cov)  # Number false positives = negatives covered by the ruleset
    fn = len(pos_idx) - len(
        pos_cov
    )  # Number false negatives = positives not covered by the ruleset
    exceptions_dl = math.log2(base_functions.nCr(p, fp)) + math.log2(
        base_functions.nCr((N - p), fn)
    )
    if verbosity >= 5:
        print(
            f"exceptions_bits| {ruleset.truncstr()}: \n N {N} p {p} fp {fp} fn {fn}: exceptions_bits {rnd(exceptions_dl)}"
        )

    return exceptions_dl


def _rs_total_bits(
    ruleset,
    possible_conds,
    pos_df,
    neg_df,
    bestsubset_dl=False,
    ret_bestsubset=False,
    verbosity=0,
):
    """Returns total description length (in bits) of ruleset -- the sum of its theory dl and exceptions dl.

    bestsubset_dl : bool, default=False
        Whether to return estimated minimum possible dl were all rules that increase dl to be removed.
    ret_bestsubset : bool, default=False
        Whether to return the best subset that was found. Return format will be (<Ruleset>,dl).
    """

    # The RIPPER paper is brief and unclear w/r how to evaluate a ruleset for best potential dl.

    # 1) Do you reevaluate already-visited rules or evaluate each rule independently of one another?
    # Weka's source code comments that you are not supposed to, and that this is "bizarre."
    # Perhaps not recursing so -- and getting a possibly sub-optimal mdl -- could be viewed as a greedy time-saver?
    # After all, this is supposed to be an iterative algorithm, it could optimize more times with future k's,
    # and it's not like we're performing an exhaustive search of every possible combination anyways.

    # 2) In what order are you supposed to evaluate? FIFO or LIFO?
    # Footnote 7 suggests optimization is done FIFO; the previous page suggests IREP* final dl reduction is done LIFO;
    # and context suggests dl reduction should be performed the same way both times.

    # I chose to greedy for #1, and FIFO for #2 but may choose differently in a future version if it seems more appropriate.
    # In any case, RIPPER's strong performance on the test sets vs. RandomForest suggests it may not matter all that much.

    # if type(ruleset) != Ruleset:
    #    raise TypeError(f'param ruleset in _rs_total_bits should be type Ruleset')
    if ret_bestsubset and not bestsubset_dl:
        raise ValueError(
            f"ret_bestsubset must be True in order to return bestsubset_dl"
        )

    if not bestsubset_dl:
        theory_bits = _rs_theory_bits(ruleset, possible_conds, verbosity=verbosity)
        data_bits = _exceptions_bits(ruleset, pos_df, neg_df, verbosity=verbosity)
        if verbosity >= 3:
            print(f"total ruleset bits | {rnd(theory_bits + data_bits)}")
        return theory_bits + data_bits
    else:
        # Collect the dl of each subset
        subset_dls = []
        theory_dl = 0
        if verbosity >= 5:
            print(f"find best potential dl for {ruleset}:")
        for i, rule in enumerate(
            ruleset.rules
        ):  # Separating theory and exceptions dls in this way means you don't have to recalculate theory each time
            subset = Ruleset(ruleset.rules[: i + 1])
            rule_theory_dl = _r_theory_bits(rule, possible_conds, verbosity=verbosity)
            theory_dl += rule_theory_dl
            exceptions_dl = _exceptions_bits(
                subset, pos_df, neg_df, verbosity=verbosity
            )
            subset_dls.append(theory_dl + exceptions_dl)
            if verbosity >= 5:
                print(f"subset 0-{i} | dl: {rnd(subset_dls[i])}")

        # Build up the best Ruleset and calculate the mdl
        mdl_ruleset = Ruleset()
        for i, rule, in enumerate(ruleset.rules):
            if (
                i == 0 or subset_dls[i] <= subset_dls[i - 1]
            ):  # Rule i does not worsen the dl
                mdl_ruleset.add(rule)
        if verbosity >= 5:
            print(f"subset dls: {[(i,rnd(dl)) for i,dl in enumerate(subset_dls)]}")
            print(f"best potential ruleset: {mdl_ruleset}")
        mdl = _rs_total_bits(
            mdl_ruleset,
            possible_conds,
            pos_df,
            neg_df,
            bestsubset_dl=False,
            verbosity=0,
        )  # About to print value below
        if verbosity >= 5:
            print(f"best potential dl was {rnd(mdl)}")
            print()
        if not ret_bestsubset:
            return mdl
        else:
            return (mdl_ruleset, mdl)


def _rs_total_bits_cn(
    cn,
    ruleset,
    possible_conds,
    pos_idx,
    neg_idx,
    bestsubset_dl=False,
    ret_bestsubset=False,
    verbosity=0,
):
    """Returns total description length (in bits) of ruleset -- the sum of its theory dl and exceptions dl.

    bestsubset_dl : bool, default=False
        Whether to return estimated minimum possible dl were all rules that increase dl to be removed.
    ret_bestsubset : bool, default=False
        Whether to return the best subset that was found. Return format will be (<Ruleset>,dl).
    """

    # The RIPPER paper is brief and unclear w/r how to evaluate a ruleset for best potential dl.

    # 1) Do you reevaluate already-visited rules or evaluate each rule independently of one another?
    # Weka's source code comments that you are not supposed to, and that this is "bizarre."
    # Perhaps not recursing so -- and getting a possibly sub-optimal mdl -- could be viewed as a greedy time-saver?
    # After all, this is supposed to be an iterative algorithm, it could optimize more times with future k's,
    # and it's not like we're performing an exhaustive search of every possible combination anyways.

    # 2) In what order are you supposed to evaluate? FIFO or LIFO?
    # Footnote 7 suggests optimization is done FIFO; the previous page suggests IREP* final dl reduction is done LIFO;
    # and context suggests dl reduction should be performed the same way both times.

    # I chose to greedy for #1, and FIFO for #2 but may choose differently in a future version if it seems more appropriate.
    # In any case, RIPPER's strong performance on the test sets vs. RandomForest suggests it may not matter all that much.

    # if type(ruleset) != Ruleset:
    #    raise TypeError(f'param ruleset in _rs_total_bits should be type Ruleset')
    if ret_bestsubset and not bestsubset_dl:
        raise ValueError(
            f"ret_bestsubset must be True in order to return bestsubset_dl"
        )

    if not bestsubset_dl:
        theory_bits = _rs_theory_bits(ruleset, possible_conds, verbosity=verbosity)
        data_bits = _exceptions_bits_cn(
            cn, ruleset, pos_idx, neg_idx, verbosity=verbosity
        )
        if verbosity >= 3:
            print(f"total ruleset bits | {rnd(theory_bits + data_bits)}")
        return theory_bits + data_bits
    else:
        # Collect the dl of each subset
        subset_dls = []
        theory_dl = 0
        if verbosity >= 5:
            print(f"find best potential dl for {ruleset}:")
        for i, rule in enumerate(
            ruleset.rules
        ):  # Separating theory and exceptions dls in this way means you don't have to recalculate theory each time
            subset = Ruleset(ruleset.rules[: i + 1])
            rule_theory_dl = _r_theory_bits(rule, possible_conds, verbosity=verbosity)
            theory_dl += rule_theory_dl
            exceptions_dl = _exceptions_bits_cn(
                cn, subset, pos_idx, neg_idx, verbosity=verbosity
            )
            subset_dls.append(theory_dl + exceptions_dl)
            if verbosity >= 5:
                print(f"subset 0-{i} | dl: {rnd(subset_dls[i])}")

        # Build up the best Ruleset and calculate the mdl
        mdl_ruleset = Ruleset()
        for i, rule, in enumerate(ruleset.rules):
            if (
                i == 0 or subset_dls[i] <= subset_dls[i - 1]
            ):  # Rule i does not worsen the dl
                mdl_ruleset.add(rule)
        if verbosity >= 5:
            print(f"subset dls: {[(i,rnd(dl)) for i,dl in enumerate(subset_dls)]}")
            print(f"best potential ruleset: {mdl_ruleset}")
        mdl = _rs_total_bits_cn(
            cn,
            mdl_ruleset,
            possible_conds,
            pos_idx,
            neg_idx,
            bestsubset_dl=False,
            verbosity=0,
        )  # About to print value below
        if verbosity >= 5:
            print(f"best potential dl was {rnd(mdl)}")
            print()
        if not ret_bestsubset:
            return mdl
        else:
            return (mdl_ruleset, mdl)



# Author: Ilan Moscovitz <ilan.moscovitz@gmail.com>
# License: MIT

import copy
from functools import reduce
import math
import numpy as np
import operator as op

import pandas as pd
from random import shuffle, seed

from wittgenstein.base import Cond, Rule, Ruleset, rnd
from wittgenstein.check import _warn

##########################
##### BASE FUNCTIONS #####
##########################


def grow_rule(
    pos_df,
    neg_df,
    possible_conds,
    initial_rule=Rule(),
    max_rule_conds=None,
    verbosity=0,
):
    """Fit a new rule to add to a ruleset"""

    rule0 = copy.deepcopy(initial_rule)
    if verbosity >= 4:
        print(f"growing rule from initial rule: {rule0}")

    rule1 = copy.deepcopy(rule0)
    while (len(rule0.covers(neg_df)) > 0 and rule1 is not None) and (
        max_rule_conds is None or len(rule1.conds) < max_rule_conds
    ):  # Stop refining rule if no negative examples remain
        rule1 = best_successor(
            rule0, possible_conds, pos_df, neg_df, verbosity=verbosity
        )
        if rule1 is not None:
            rule0 = rule1
            if verbosity >= 4:
                print(f"negs remaining {len(rule0.covers(neg_df))}")

    if not rule0.isempty():
        if verbosity >= 2:
            print(f"grew rule: {rule0}")
        return rule0
    else:
        # warning_str = f"grew an empty rule: {rule0} over {len(pos_idx)} pos and {len(neg_idx)} neg"
        # _warn(warning_str, RuntimeWarning, filename='base_functions', funcname='grow_rule')
        return rule0


def grow_rule_cn(
    cn, pos_idx, neg_idx, initial_rule=Rule(), max_rule_conds=None, verbosity=0
):
    """Fit a new rule to add to a ruleset. (Optimized version.)"""

    rule0 = copy.deepcopy(initial_rule)
    rule1 = copy.deepcopy(rule0)
    if verbosity >= 4:
        print(f"growing rule from initial rule: {rule0}")

    num_neg_covered = len(cn.rule_covers(rule0, subset=neg_idx))
    while num_neg_covered > 0:  # Stop refining rule if no negative examples remain
        user_halt = max_rule_conds is not None and len(rule1.conds) >= max_rule_conds
        if user_halt:
            break

        rule1 = best_rule_successor_cn(cn, rule0, pos_idx, neg_idx, verbosity=verbosity)
        if rule1 is None:
            break
        rule0 = rule1
        num_neg_covered = len(cn.rule_covers(rule0, neg_idx))
        if verbosity >= 4:
            print(f"negs remaining: {num_neg_covered}")

    if not rule0.isempty():
        if verbosity >= 2:
            print(f"grew rule: {rule0}")
        return rule0
    else:
        # warning_str = f"grew an empty rule: {rule0} over {len(pos_idx)} pos and {len(neg_idx)} neg"
        # _warn(warning_str, RuntimeWarning, filename='base_functions', funcname='grow_rule_cn')
        return rule0


def prune_rule(
    rule,
    prune_metric,
    pos_pruneset,
    neg_pruneset,
    eval_index_on_ruleset=None,
    verbosity=0,
):
    """Return a pruned version of the Rule by removing Conds.

    rule : Rule
        Rule to prune.
    prune_metric : function
        Function that returns a value to maximize.
    pos_pruneset : DataFrame
        Positive class examples.
    neg_pruneset : DataFrame
        Negative class examples.

    eval_index_on_ruleset : tuple(rule_index, Ruleset), default=None
        Pass the rest of the Rule's Ruleset (excluding the Rule in question),
        in order to prune the rule based on the performance of its entire Ruleset,
        rather than on the rule alone. Used during optimization stage of RIPPER.
    verbosity : int (0-5), default=0
        Output verbosity.
    """

    if rule.isempty():
        # warning_str = f"can't prune empty rule: {rule}"
        # _warn(warning_str, RuntimeWarning, filename='base_functions', funcname='prune_rule')
        return rule

    if not eval_index_on_ruleset:

        # Currently-best pruned rule and its prune value
        best_rule = copy.deepcopy(rule)
        best_v = 0

        # Iterative test rule
        current_rule = copy.deepcopy(rule)

        while current_rule.conds:
            v = prune_metric(current_rule, pos_pruneset, neg_pruneset)
            if verbosity >= 5:
                print(f"prune value of {current_rule}: {rnd(v)}")
            if v is None:
                return None
            if v >= best_v:
                best_v = v
                best_rule = copy.deepcopy(current_rule)
            current_rule.conds.pop(-1)

        if verbosity >= 2:
            if len(best_rule.conds) != len(rule.conds):
                print(f"pruned rule: {best_rule}")
            else:
                print(f"pruned rule unchanged")
        return best_rule

    else:
        # Check if index matches rule to prune
        rule_index, ruleset = eval_index_on_ruleset
        if ruleset.rules[rule_index] != rule:
            raise ValueError(
                f"rule mismatch: {rule} - {ruleset.rules[rule_index]} in {ruleset}"
            )

        current_ruleset = copy.deepcopy(ruleset)
        current_rule = current_ruleset.rules[rule_index]
        best_ruleset = copy.deepcopy(current_ruleset)
        best_v = 0

        # Iteratively prune and test rule over ruleset.
        # This is unfortunately expensive.
        while current_rule.conds:
            v = prune_metric(current_ruleset, pos_pruneset, neg_pruneset)
            if verbosity >= 5:
                print(f"prune value of {current_rule}: {rnd(v)}")
            if v is None:
                return None
            if v >= best_v:
                best_v = v
                best_rule = copy.deepcopy(current_rule)
                best_ruleset = copy.deepcopy(current_ruleset)
            current_rule.conds.pop(-1)
            current_ruleset.rules[rule_index] = current_rule
        return best_rule


def prune_rule_cn(
    cn, rule, prune_metric_cn, pos_idx, neg_idx, eval_index_on_ruleset=None, verbosity=0
):
    """Return a pruned version of the Rule by removing Conds. (Optimized version.)

    rule : Rule
        Rule to prune.
    prune_metric : function
        Function that returns a value to maximize.
    pos_pruneset : DataFrame
        Positive class examples.
    neg_pruneset : DataFrame
        Negative class examples.

    eval_index_on_ruleset : tuple(rule_index, Ruleset), default=None
        Pass the rest of the Rule's Ruleset (excluding the Rule in question),
        in order to prune the rule based on the performance of its entire Ruleset,
        rather than on the rule alone. Used during optimization stage of RIPPER.
    verbosity : int (0-5), default=0
        Output verbosity.
    """

    if rule.isempty():
        # warning_str = f"can't prune empty rule: {rule}"
        # _warn(warning_str, RuntimeWarning, filename='base_functions', funcname='prune_rule_cn')
        return rule

    if not eval_index_on_ruleset:

        # Currently-best pruned rule and its prune value
        best_rule = copy.deepcopy(rule)
        best_v = 0

        # Iterative test rule
        current_rule = copy.deepcopy(rule)

        while current_rule.conds:
            v = prune_metric_cn(cn, current_rule, pos_idx, neg_idx)
            if verbosity >= 5:
                print(f"prune value of {current_rule}: {rnd(v)}")
            if v is None:
                return None
            if v >= best_v:
                best_v = v
                best_rule = copy.deepcopy(current_rule)
            current_rule.conds.pop(-1)

        if verbosity >= 2:
            if len(best_rule.conds) != len(rule.conds):
                print(f"pruned rule: {best_rule}")
            else:
                print(f"pruned rule unchanged")
        return best_rule

    # cn is Untouched below here
    else:
        # Check if index matches rule to prune
        rule_index, ruleset = eval_index_on_ruleset
        if ruleset.rules[rule_index] != rule:
            raise ValueError(
                f"rule mismatch: {rule} - {ruleset.rules[rule_index]} in {ruleset}"
            )

        current_ruleset = copy.deepcopy(ruleset)
        current_rule = current_ruleset.rules[rule_index]
        best_ruleset = copy.deepcopy(current_ruleset)
        best_v = 0

        # Iteratively prune and test rule over ruleset.
        while current_rule.conds:
            v = prune_metric_cn(cn, current_rule, pos_idx, neg_idx)
            if verbosity >= 5:
                print(f"prune value of {current_rule}: {rnd(v)}")
            if v is None:
                return None
            if v >= best_v:
                best_v = v
                best_rule = copy.deepcopy(current_rule)
                best_ruleset = copy.deepcopy(current_ruleset)
            current_rule.conds.pop(-1)
            current_ruleset.rules[rule_index] = current_rule
        return best_rule


def recalibrate_proba(
    ruleset, Xy_df, class_feat, pos_class, min_samples=10, require_min_samples=True
):
    """Recalibrate a Ruleset's probability estimations using unseen labeled data without changing the underlying model. May improve .predict_proba generalizability.
    Does not affect the underlying model or which predictions it makes -- only probability estimates. Use params min_samples and require_min_samples to select desired behavior.

    Note1: RunTimeWarning will occur as a reminder when min_samples and require_min_samples params might result in unintended effects.
    Note2: It is possible recalibrating could result in some positive .predict predictions with <0.5 .predict_proba positive probability.

    ruleset : Ruleset
        Ruleset to recalibrate.
    Xy_df : DataFrame
        Labeled dataset.
    class_feat : str
        Name of class feature column in Xy_df.
    pos_class : value, typically str or int
        Positive class value.

    min_samples : int, default=10
        Required minimum number of samples per Rule. Regardless of min_samples, at least one sample of the correct class is always required.
    require_min_samples : bool, default=True
        Halt (with warning) if any Rule lacks the minimum number of samples.
        Setting to False will warn, but still replace Rules probabilities even if the minimum number of samples is not met.
    """

    # At least this many samples per rule (or neg) must be of correct class
    required_correct_samples = 1

    # If not using min_samples, set it to 1
    if not min_samples or min_samples < 1:
        min_samples = 1

    # Collect each Rule's pos and neg frequencies in list "rule_class_freqs"
    # Store rules that lack enough samples in list "insufficient_rules"
    df = Xy_df

    rule_class_freqs = [None] * len(ruleset.rules)
    insufficient_rules = []
    for i, rule in enumerate(ruleset.rules):
        npos_pred = num_pos(rule.covers(df), class_feat=class_feat, pos_class=pos_class)
        nneg_pred = num_neg(rule.covers(df), class_feat=class_feat, pos_class=pos_class)
        neg_pos_pred = (nneg_pred, npos_pred)
        rule_class_freqs[i] = neg_pos_pred
        # Rule has insufficient samples if fewer than minsamples or lacks at least one correct sample
        if (
            sum(neg_pos_pred) < min_samples
            or sum(neg_pos_pred) < 1
            or neg_pos_pred[0] < required_correct_samples
        ):
            insufficient_rules.append(rule)

    # Collect class frequencies for negative predictions
    uncovered = df.drop(ruleset.covers(df).index)
    neg_freq = num_neg(uncovered, class_feat=class_feat, pos_class=pos_class)
    tn_fn = (neg_freq, len(uncovered) - neg_freq)

    # Issue warnings if trouble with sample size
    if require_min_samples:
        if insufficient_rules:  # WARN if/which rules lack enough samples
            pretty_insufficient_rules = "\n".join([str(r) for r in insufficient_rules])
            warning_str = f"param min_samples={min_samples}; insufficient number of samples or fewer than {required_correct_samples} correct samples for rules {pretty_insufficient_rules}"
            _warn(
                warning_str,
                RuntimeWarning,
                filename="base_functions",
                funcname="recalibrate_proba",
            )
        if neg_freq < min_samples or tn_fn[1] < 1:  # WARN if neg lacks enough samples
            warning_str = f"param min_samples={min_samples}; insufficient number of negatively labled samples"
            _warn(
                warning_str,
                RuntimeWarning,
                filename="base_functions",
                funcname="recalibrate_proba",
            )
        if insufficient_rules or sum(tn_fn) < min_samples:
            if (
                require_min_samples
            ):  # WARN if require_min_samples -> halting recalibration
                warning_str = f"Recalibrating halted. to recalibrate, try using more samples, lowering min_samples, or set require_min_samples to False"
                _warn(
                    warning_str,
                    RuntimeWarning,
                    filename="base_functions",
                    funcname="recalibrate_proba",
                )
                return
            else:  # GO AHEAD EVEN THOUGH NOT ENOUGH SAMPLES
                pass
                # warning_str = f'Because require_min_samples=False, recalibrating probabilities for any rules with enough samples min_samples>={min_samples} that have at least {required_correct_samples} correct samples even though not all rules have enough samples. Probabilities for any rules that lack enough samples will be retained.'
                # _warn(warning_str, RuntimeWarning, filename='base_functions', funcname='recalibrate_proba')

    # Assign collected frequencies to Rules
    for rule, freqs in zip(ruleset.rules, rule_class_freqs):
        if sum(freqs) >= min_samples and freqs[0] >= required_correct_samples:
            rule.class_freqs = freqs
        else:
            rule.class_freqs = None

    # Assign Ruleset's uncovered frequencies
    if not hasattr(ruleset, "uncovered_class_freqs") or (
        neg_freq >= min_samples and tn_fn[1] >= required_correct_samples
    ):
        ruleset.uncovered_class_freqs = tn_fn

    # Warn if no pos or no neg samples
    noneg = (
        sum([freqs[0] for freqs in rule_class_freqs]) + ruleset.uncovered_class_freqs[0]
    ) == 0
    nopos = (
        sum([freqs[1] for freqs in rule_class_freqs]) + ruleset.uncovered_class_freqs[1]
    ) == 0
    if noneg:
        warning_str = f"No negative samples."
        _warn(
            warning_str,
            RuntimeWarning,
            filename="base_functions",
            funcname="recalibrate_proba",
        )
    if nopos:
        warning_str = f"No positive samples"
        _warn(
            warning_str,
            RuntimeWarning,
            filename="base_functions",
            funcname="recalibrate_proba",
        )

    ###################
    ##### METRICS #####
    ###################


def gain(before, after, pos_df, neg_df):
    """Calculates the information gain from before to after."""
    p0count = before.num_covered(pos_df)
    p1count = after.num_covered(pos_df)
    n0count = before.num_covered(neg_df)
    n1count = after.num_covered(neg_df)
    return p1count * (
        math.log2((p1count + 1) / (p1count + n1count + 1))
        - math.log2((p0count + 1) / (p0count + n0count + 1))
    )


def gain_cn(cn, cond_step, rule_covers_pos_idx, rule_covers_neg_idx):
    """Calculates the information gain from adding a Cond."""
    p0count = len(rule_covers_pos_idx)
    p1count = len(cn.cond_covers(cond_step, subset=rule_covers_pos_idx))
    n0count = len(rule_covers_neg_idx)
    n1count = len(cn.cond_covers(cond_step, subset=rule_covers_neg_idx))
    return p1count * (
        math.log2((p1count + 1) / (p1count + n1count + 1))
        - math.log2((p0count + 1) / (p0count + n0count + 1))
    )


def precision(object, pos_df, neg_df):
    """Calculate precision value of object's classification.

    object : Cond, Rule, or Ruleset
    """

    pos_covered = object.covers(pos_df)
    neg_covered = object.covers(neg_df)
    total_n_covered = len(pos_covered) + len(neg_covered)
    if total_n_covered == 0:
        return None
    else:
        return len(pos_covered) / total_n_covered


def rule_precision_cn(cn, rule, pos_idx, neg_idx):
    """Calculate precision value of object's classification.

    object : Cond, Rule, or Ruleset
    """

    pos_covered = cn.rule_covers(rule, pos_idx)
    neg_covered = cn.rule_covers(rule, neg_idx)
    total_n_covered = len(pos_covered) + len(neg_covered)
    if total_n_covered == 0:
        return None
    else:
        return len(pos_covered) / total_n_covered


def score_accuracy(predictions, actuals):
    """Calculate accuracy score of a trained model on a test set.

    predictions : iterable<bool>
        True for predicted positive class, False otherwise.
    actuals : iterable<bool>
        True for actual positive class, False otherwise.
    """
    t = [pr for pr, act in zip(predictions, actuals) if pr == act]
    n = predictions
    return len(t) / len(n)


def _accuracy(object, pos_pruneset, neg_pruneset):
    """Calculate accuracy value of object's classification.

    object : Cond, Rule, or Ruleset
    """
    P = len(pos_pruneset)
    N = len(neg_pruneset)
    if P + N == 0:
        return None

    tp = len(object.covers(pos_pruneset))
    tn = N - len(object.covers(neg_pruneset))
    return (tp + tn) / (P + N)


def _rule_accuracy_cn(cn, rule, pos_pruneset_idx, neg_pruneset_idx):
    """Calculate accuracy value of object's classification.

    object: Cond, Rule, or Ruleset
    """
    P = len(pos_pruneset_idx)
    N = len(neg_pruneset_idx)
    if P + N == 0:
        return None

    tp = len(cn.rule_covers(rule, pos_pruneset_idx))
    tn = N - len(cn.rule_covers(rule, neg_pruneset_idx))
    return (tp + tn) / (P + N)


def best_successor(rule, possible_conds, pos_df, neg_df, verbosity=0):
    """Return for a Rule its best successor Rule according to FOIL information gain metric."""

    best_gain = 0
    best_successor_rule = None

    for successor in rule.successors(possible_conds, pos_df, neg_df):
        g = gain(rule, successor, pos_df, neg_df)
        if g > best_gain:
            best_gain = g
            best_successor_rule = successor
    if verbosity >= 5:
        print(f"gain {rnd(best_gain)} {best_successor_rule}")
    return best_successor_rule


def best_rule_successor_cn(cn, rule, pos_idx, neg_idx, verbosity=0):
    """Return for a Rule its best successor Rule according to FOIL information gain metric."""

    best_cond = None
    best_gain = float("-inf")

    rule_covers_pos_idx = cn.rule_covers(rule, pos_idx)
    rule_covers_neg_idx = cn.rule_covers(rule, neg_idx)

    for cond_action_step in cn.conds:
        g = gain_cn(cn, cond_action_step, rule_covers_pos_idx, rule_covers_neg_idx)
        if g > best_gain:
            best_gain = g
            best_cond = cond_action_step
    if verbosity >= 5:
        print(f"gain {rnd(best_gain)} {best_cond}")
    return Rule(rule.conds + [best_cond]) if best_gain > 0 else None


###################
##### HELPERS #####
###################


def pos_neg_split(df, class_feat, pos_class):
    """Split df into pos and neg classes."""
    pos_df = pos(df, class_feat, pos_class)
    neg_df = neg(df, class_feat, pos_class)
    return pos_df, neg_df


def df_shuffled_split(df, split_size=0.66, random_state=None):
    """Return tuple of shuffled and split DataFrame.

    split_size : float
        Proportion of rows to include in return[0].
    random_state : float, default=None
        Random seed.

    Returns
        Tuple of shuffled and split DataFrame.
    """
    idx1, idx2 = random_split(
        df.index, split_size, res_type=set, random_state=random_state
    )
    return df.loc[idx1, :], df.loc[idx2, :]


def set_shuffled_split(set_to_split, split_size, random_state=None):
    """Return tuple of shuffled and split set.

    split_size : float
        Proportion of set to include in return[0].
    random_state : float, default=None
        Random seed.

    Returns
        Tuple of shuffled and split DataFrame.
    """
    list_to_split = list(set_to_split)
    seed(random_state)
    shuffle(list_to_split)
    split_at = int(len(list_to_split) * split_size)
    return (set(list_to_split[:split_at]), set(list_to_split[split_at:]))


def random_split(to_split, split_size, res_type=set, random_state=None):
    """Return tuple of shuffled and split iterable.

    to_split : iterable
        What to shuffle and split.
    split_size : float
        Proportion to include in return[0].
    res_type : type
        Type of items to return.
    random_state : float, default=None
        Random seed.
    Returns
        Tuple of shuffled and split DataFrame.
    """
    to_split = list(to_split)
    seed(random_state)
    shuffle(to_split)
    split_at = int(len(to_split) * split_size)
    return (res_type(to_split[:split_at]), res_type(to_split[split_at:]))


def pos(df, class_feat, pos_class):
    """Return subset of instances that are labeled positive."""
    return df[df[class_feat] == pos_class]


def neg(df, class_feat, pos_class):
    """Return subset of instances that are labeled negative."""
    return df[df[class_feat] != pos_class]


def num_pos(df, class_feat, pos_class):
    """Return number of instances that are labeled positive."""
    return len(df[df[class_feat] == pos_class])


def num_neg(df, class_feat, pos_class):
    """ Return number of instances that are labeled negative."""
    return len(df[df[class_feat] != pos_class])


def nCr(n, r):
    """Return number of combinations C(n, r)."""

    def product(numbers):
        return reduce(op.mul, numbers, 1)

    num = product(range(n, n - r, -1))
    den = product(range(1, r + 1))
    return num // den


def argmin(iterable):
    """Return index of minimum value."""
    lowest_val = iterable[0]
    lowest_i = 0
    for i, val in enumerate(iterable):
        if val < lowest_val:
            lowest_val = val
            lowest_i = i
    return lowest_i


def i_replaced(list_, i, value):
    """Return a new list with element i replaced by value.

    i : value
        Index to replace with value.
    value : value
        Value to replace at index i. None will return original list with element i removed.
    """
    if value is not None:
        return list_[:i] + [value] + list_[i + 1 :]
    else:
        return list_[:i] + list_[i + 1 :]


def rm_covered(object, pos_df, neg_df):
    """Return pos and neg dfs of examples that are not covered by object.

    Parameters
    ----------
    object : Cond, Rule, or Ruleset
        Object whose coverage predictions to invoke.
    pos_df : DataFrame
        Positive examples.
    neg_df : DataFrame
        Negative examples.

    Return
    ------
    tuple<DataFrame>
        Positive and negative examples not covered by object.
    """
    return (
        pos_df.drop(object.covers(pos_df).index, axis=0, inplace=False),
        neg_df.drop(object.covers(neg_df).index, axis=0, inplace=False),
    )


def rm_rule_covers_cn(cn, rule, pos_idx, neg_idx):
    """Return positive and negative indices not covered by object."""
    return (
        pos_idx - cn.rule_covers(rule, pos_idx),
        neg_idx - cn.rule_covers(rule, neg_idx),
    )


def aslist(data):
    try:
        return data.aslist()
    except:
        return data


def truncstr(iterable, limit=5, direction="left"):
    """Return Ruleset string representation limited to a specified number of rules.

    limit: how many rules to return
    direction: which part to return. (valid options: 'left', 'right')
    """
    if len(iterable) > limit:
        if direction == "left":
            return iterable[:limit].__str__() + "..."
        elif direction == "right":
            return "..." + iterable[-limit:].__str__()
        else:
            raise ValueError('direction param must be "left" or "right"')
    else:
        return str(iterable)


def stop_early(ruleset, max_rules, max_total_conds):
    """Function to decide whether to halt training."""
    return (max_rules is not None and len(ruleset.rules) >= max_rules) or (
        max_total_conds is not None and ruleset.count_conds() >= max_total_conds
    )



import numpy as np
import pandas as pd
import pytest

from wittgenstein import IREP, RIPPER
from wittgenstein.base_functions import df_shuffled_split
from wittgenstein.base import Cond, Rule

DF = pd.read_csv("credit.csv")
CLASS_FEAT = "Class"
DEFAULT_CLASS_FEAT = "Class"
CREDIT_POS_CLASS = "+"
SPLIT_SIZE = 0.6

X_DF = DF.drop(CLASS_FEAT, axis=1)
Y_DF = DF[CLASS_FEAT]

XY_NP = DF.values
X_NP = X_DF.values
Y_NP = Y_DF.values
NP_CLASS_FEAT = -1

######

irep = IREP(random_state=42)
rip = RIPPER(random_state=42)

#####

train, test = df_shuffled_split(DF, random_state=42)
test_x, test_y = test.drop(CLASS_FEAT, axis=1), test[CLASS_FEAT]

irep.fit(train, class_feat=CLASS_FEAT, pos_class=CREDIT_POS_CLASS)
rip.fit(train, class_feat=CLASS_FEAT, pos_class=CREDIT_POS_CLASS)

#####


def test_predict():
    irep_preds = irep.predict(test_x)
    assert all(p in (True, False) for p in irep_preds)
    assert not all(p == True for p in irep_preds)
    assert not all(p == False for p in irep_preds)

    rip_preds = rip.predict(test_x)
    assert all(p in (True, False) for p in rip_preds)
    assert not all(p == True for p in rip_preds)
    assert not all(p == False for p in rip_preds)


def test_predict_give_reasons():
    def reason_iff_pos(pred, reason):
        if list(reason) != []:
            return pred
        else:
            return not pred

    irep_preds = irep.predict(test_x, give_reasons=True)
    assert all(reason_iff_pos(p, r) for p, r in zip(*irep_preds))
    rip_preds = rip.predict(test_x, give_reasons=True)
    assert all(reason_iff_pos(p, r) for p, r in zip(*rip_preds))


def test_predict_proba():
    # The vast majority of predictions and predict probas should match
    def pred_proba_match(pred, proba):
        if pred:
            return proba[1] >= proba[0]
        else:
            return proba[0] >= proba[1]

    irep_preds = irep.predict(test_x)
    irep_probas = irep.predict_proba(test_x)
    assert (
        np.mean(
            [
                pred_proba_match(pred, proba)
                for pred, proba in zip(irep_preds, irep_probas)
            ]
        )
        >= 0.95
    )
    rip_preds = rip.predict(test_x)
    rip_probas = irep.predict_proba(test_x)
    assert (
        np.mean(
            [
                pred_proba_match(pred, proba)
                for pred, proba in zip(rip_preds, irep_probas)
            ]
        )
        >= 0.95
    )


def test_score():
    assert irep.score(test_x, test_y) >= 0.75
    assert rip.score(test_x, test_y) >= 0.75



import pytest
import pandas as pd

from wittgenstein.base_functions import (
    pos,
    neg,
    pos_neg_split,
    df_shuffled_split,
    num_pos,
    num_neg,
    df_shuffled_split,
    set_shuffled_split,
    random_split,
)
from wittgenstein.base import Cond, Rule, Ruleset

FULL_DF = pd.read_csv("house-votes-84.csv")
FIRST_10_EXAMPLES = FULL_DF.head(10)
CLASS_FEAT = "Party"
POS_CLASS = "democrat"
SPLIT_SIZE = 0.6

FEATURE1, VALUE1 = "Handicapped-infants", "n"
FEATURE2, VALUE2 = "Water-project-cost-sharing", "y"
FEATURE3, VALUE3 = "anti-satellite-test-ban", "n"
FEATURE4, VALUE4 = "el-salvador-aid", "y"
COND1 = Cond(FEATURE1, VALUE1)
COND2 = Cond(FEATURE2, VALUE2)
COND3 = Cond(FEATURE3, VALUE3)
COND4 = Cond(FEATURE4, VALUE4)

##### Test Helpers ######


def test_len_pos_df_is_6():
    assert len(pos(FIRST_10_EXAMPLES, CLASS_FEAT, POS_CLASS)) == 6


def test_len_neg_df_is_4():
    assert len(neg(FIRST_10_EXAMPLES, CLASS_FEAT, POS_CLASS)) == 4


def test_split_pos_is_6_neg_is_4():
    pos_df, neg_df = pos_neg_split(FIRST_10_EXAMPLES, CLASS_FEAT, POS_CLASS)
    assert (len(pos_df), len(neg_df)) == (6, 4)


def test_shuffled_splits_are_len_7_len_3():
    df1, df2 = df_shuffled_split(FIRST_10_EXAMPLES, 0.7, random_state=None)
    assert (len(df1), len(df2)) == (7, 3)


def test_num_pos_is_6_num_neg_is_4():
    p = num_pos(FIRST_10_EXAMPLES, CLASS_FEAT, POS_CLASS)
    n = num_neg(FIRST_10_EXAMPLES, CLASS_FEAT, POS_CLASS)
    assert (p, n) == (6, 4)


def test_num_pos_is_0_num_neg_is_0():
    pos_df, neg_df = pos_neg_split(FIRST_10_EXAMPLES, CLASS_FEAT, POS_CLASS)
    p = num_pos(neg_df, CLASS_FEAT, POS_CLASS)
    n = num_neg(pos_df, CLASS_FEAT, POS_CLASS)
    assert (p, n) == (0, 0)


####### Test Conds ######


def test_conds_are_equal():
    assert Cond("feature", "value") == Cond("feature", "value")


def test_conds_are_unequal_value():
    assert Cond("feature", "value1") != Cond("feature", "value2")


def test_conds_are_unequal_feature():
    assert Cond("feature1", "value") != Cond("feature2", "value")


def test_cond_covers():
    cond = Cond(FEATURE1, VALUE1)
    assert set(cond.covers(FIRST_10_EXAMPLES)[FEATURE1].tolist()) == {VALUE1}


def test_cond_doesnt_cover():
    cond = Cond(FEATURE1, VALUE1)
    assert (
        len(cond.covers(FIRST_10_EXAMPLES[FIRST_10_EXAMPLES[FEATURE1] != VALUE1])) == 0
    )


def test_cond_num_covered_is_7():
    cond = Cond(FEATURE1, VALUE1)
    assert len(cond.covers(FIRST_10_EXAMPLES)[FEATURE1].tolist()) == 7


##### Test Rule #####


def test_empty_rules_equal():
    assert Rule() == Rule()


def test_rules_len_1_equal():
    assert Rule([COND1]) == Rule([COND1])


def test_rules_len_1_unequal():
    assert Rule([COND1]) != Rule([COND2])


def test_rules_len_2_equal():
    rule1 = Rule([COND1, COND2])
    rule2 = Rule([COND1, COND2])
    assert rule1 == rule2


def test_disordered_rules_len_2_equal():
    rule1 = Rule([COND1, COND2])
    rule2 = Rule([COND2, COND1])
    assert rule1 == rule2


def test_rules_len_2_unequal():
    rule1 = Rule([COND1, COND2])
    rule2 = Rule([COND2, COND3])
    assert rule1 != rule2


def test_rule_covers():
    rule = Rule([COND1, COND4])
    rule_covers = rule.covers(FIRST_10_EXAMPLES)
    assertion_df = FIRST_10_EXAMPLES[
        (FIRST_10_EXAMPLES[FEATURE1] == VALUE1)
        & (FIRST_10_EXAMPLES[FEATURE4] == VALUE4)
    ]
    assert set(rule_covers.index.tolist()) == set(assertion_df.index.tolist())


def test_rule_num_covered():
    rule = Rule([COND1, COND4])
    rule_num_covered = rule.num_covered(FIRST_10_EXAMPLES)
    len_assertion_df = len(
        FIRST_10_EXAMPLES[
            (FIRST_10_EXAMPLES[FEATURE1] == VALUE1)
            & (FIRST_10_EXAMPLES[FEATURE4] == VALUE4)
        ]
    )
    assert rule_num_covered == len_assertion_df


#### Functions

### Ensure random states behave consistently:


def test_random_split_set_shuffled_split_are_same():
    ssp1, ssp2 = set_shuffled_split(range(len(FULL_DF)), 0.66, random_state=42)
    rs1, rs2 = random_split(range(len(FULL_DF)), 0.66, res_type=list, random_state=42)
    assert set(rs1) == set(ssp1)
    assert set(rs2) == set(ssp2)


def test_random_split_df_shuffled_split_are_same():
    idx1, idx2 = random_split(FULL_DF.index, 0.66, res_type=set, random_state=42)
    df1, df2 = df_shuffled_split(FULL_DF, split_size=0.66, random_state=42)
    assert (set(FULL_DF.loc[idx1, :].index), set(FULL_DF.loc[idx2, :].index)) == (
        idx1,
        idx2,
    )



import numpy as np

import pandas as pd

from wittgenstein.base_functions import truncstr, rnd
from wittgenstein.check import (
    _check_any_datasets_not_empty,
    _check_model_features_present,
)


def preprocess_training_data(preprocess_params):

    # Get params
    trainset = preprocess_params["trainset"]
    y = preprocess_params["y"]
    class_feat = preprocess_params["class_feat"]
    pos_class = preprocess_params["pos_class"]
    user_requested_feature_names = preprocess_params["feature_names"]
    n_discretize_bins = preprocess_params["n_discretize_bins"]
    verbosity = preprocess_params["verbosity"]

    # Error check
    _check_valid_input_data(
        trainset,
        y,
        class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Determine class_feat
    class_feat = _get_class_feat_name(class_feat, y)

    # Build new DataFrame containing both X and y.
    df = _convert_to_training_df(
        trainset,
        y,
        class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Define pos_class
    pos_class = _get_pos_class(df, class_feat, pos_class)

    # Infer correct datatypes
    df = df.infer_objects()

    # Bin, if necessary
    df, bin_transformer_ = _try_bin_fit_or_fittransform_(
        df,
        n_discretize_bins=n_discretize_bins,
        ignore_feats=[class_feat],
        verbosity=verbosity,
    )

    # Done
    return df, class_feat, pos_class, bin_transformer_


def preprocess_prediction_data(preprocess_params):

    X = preprocess_params["X"]
    class_feat = preprocess_params["class_feat"]
    pos_class = preprocess_params["pos_class"]
    user_requested_feature_names = preprocess_params["user_requested_feature_names"]
    selected_features_ = preprocess_params["selected_features_"]
    trainset_features_ = preprocess_params["trainset_features_"]
    bin_transformer_ = preprocess_params["bin_transformer_"]
    verbosity = preprocess_params["verbosity"]

    # Error check
    _check_valid_input_data(
        X,
        y=None,
        class_feat=class_feat,
        requires_label=False,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Build new DataFrame containing both X and y.
    df = _convert_to_prediction_df(
        X,
        class_feat=class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Make sure selected features are present
    _check_model_features_present(df, selected_features_)

    # Infer correct datatypes
    df = df.infer_objects()

    # Bin, if necessary
    df, _ = _try_bin_fit_or_fittransform_(
        df,
        bin_transformer_=bin_transformer_,
        ignore_feats=[class_feat],
        verbosity=verbosity,
    )

    # Done
    return df


def _preprocess_recalibrate_proba_data(preprocess_params):

    # Get params
    X_or_Xy = preprocess_params["X_or_Xy"]
    y = preprocess_params["y"]
    class_feat = preprocess_params["class_feat"]
    pos_class = preprocess_params["pos_class"]
    user_requested_feature_names = preprocess_params["user_requested_feature_names"]
    bin_transformer_ = preprocess_params["bin_transformer_"]
    verbosity = preprocess_params["verbosity"]

    # Error check
    _check_valid_input_data(
        X_or_Xy,
        y,
        class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Build new DataFrame containing both X and y.
    df = _convert_to_training_df(
        X_or_Xy,
        y,
        class_feat,
        user_requested_feature_names=user_requested_feature_names,
    )

    # Infer correct datatypes
    df = df.infer_objects()

    # Bin, if necessary
    df, bin_transformer_ = _try_bin_fit_or_fittransform_(
        df,
        bin_transformer_=bin_transformer_,
        ignore_feats=[class_feat],
        verbosity=verbosity,
    )

    # Done
    return df


def _preprocess_y_score_data(y):
    """Return python iterable of y values."""

    def raise_wrong_ndim():
        raise IndexError(f"y input data has wrong number dimensions. It should have 1.")

    # If it's pandas or np...
    if hasattr(y, "ndim"):
        if y.ndim == 1:
            return y.tolist()
        else:
            raise_wrong_ndim()

    # Otherwise try for python iterable
    if hasattr(y, "__getitem__"):
        # No super clean way to check for 1D, but this should be pretty good
        if any([hasattr(item, "__getitem__") and type(item) is not str for item in y]):
            raise_wrong_ndim()
        else:
            return y

    # Otherwise, no idea what's going on
    else:
        raise TypeError(
            f"Could not identify valid type for y input data: {type(y)}. Recommended types are 1D python iterable, pandas Series, or 1D numpy array."
        )


def _check_valid_input_data(
    X_or_Xy,
    y=None,
    class_feat=None,
    user_requested_feature_names=None,
    requires_label=True,
):

    # Make sure there is data
    if not _check_any_datasets_not_empty([X_or_Xy]):
        raise ValueError("No data provided!")

    # If labels aren't needed, we're done
    if not requires_label:
        return

    # Ensure class feature is provided
    if (y is None) and (class_feat is None):
        raise ValueError("y or class_feat param is required")

    # Ensure target data exists if class feat is provided
    if y is None:
        if user_requested_feature_names is not None:
            feature_names = user_requested_feature_names
        elif hasattr(X_or_Xy, "columns"):
            feature_names = X_or_Xy.columns
        else:
            feature_names = list(range(len(X_or_Xy[0])))

        if class_feat not in feature_names:
            raise IndexError(
                f"Dataset does not include class feature name {class_feat}. Training set features: {feature_names}"
            )

    # If both y and class_feat provided, ensure no name mismatch between them.
    if (
        y is not None
        and class_feat is not None
        and hasattr(y, "name")
        and y.name != class_feat
    ):
        raise NameError(
            f"Feature name mismatch between params y {y.name} and class_feat {class_feat}. Besides, you only need to provide one of them."
        )


def _convert_to_training_df(X_or_Xy, y, class_feat, user_requested_feature_names=None):
    """Make a labeled Xy DataFrame from data."""

    # Create df from X_or_Xy
    if isinstance(X_or_Xy, pd.DataFrame):
        df = X_or_Xy.copy()
    else:
        df = pd.DataFrame(X_or_Xy)

    # Set feature names
    if user_requested_feature_names is not None:
        df.columns = list(
            user_requested_feature_names
        )  # list in case the type is df.columns or something

    # If necessary, merge y into df
    if y is not None:
        # If y is pd or np type, add it
        try:
            df = df.set_index(y.index)
            df[class_feat] = y.copy()
        # If that doesn't work, it's likely a python iterable
        except:
            df[class_feat] = y
    return df


def _convert_to_prediction_df(X_or_Xy, class_feat, user_requested_feature_names=None):
    """Make a labeled X DataFrame from data."""

    # Create df from X_or_Xy
    if isinstance(X_or_Xy, pd.DataFrame):
        df = X_or_Xy.copy()
    else:
        df = pd.DataFrame(X_or_Xy)

    # Drop class feature if present
    if class_feat in df.columns:
        df.drop(class_feat, axis=1, inplace=True)

    # Set feature names
    if user_requested_feature_names:
        df.columns = [f for f in user_requested_feature_names if not f == class_feat]
    return df


def _get_pos_class(df, class_feat, pos_class):
    """Get or infer the positive class name."""
    # Pos class already known
    if pos_class is not None:
        return pos_class

    # Check if pos class can be inferred as True or 1
    class_values = sorted(df[class_feat].unique())
    if len(class_values) == 2:
        # Convert numpy int64 to native python int
        try:
            class_values = [class_values[0].item(), class_values[1].item()]
        except:
            pass
        if class_values[0] is 0 and class_values[1] is 1:
            return 1
        elif class_values[0] is False and class_values[1] is True:
            return True

    # Can't infer classes
    raise NameError(
        f"Couldn't infer name of positive target class from class feature: {class_feat}. Try using parameter pos_class to specify which class label should be treated as positive, or renaming your classes as booleans or 0, 1."
    )


def _get_class_feat_name(class_feat, y):

    if class_feat is not None:
        return class_feat

    if y is not None and hasattr(y, "name"):
        # If y is a pandas Series, try to get its name
        class_feat = y.name
    else:
        # Create a name for it
        class_feat = "Class"

    return class_feat


def _try_bin_fit_or_fittransform_(
    df, ignore_feats=[], n_discretize_bins=None, bin_transformer_=None, verbosity=0
):

    # Binning has already been fit
    if bin_transformer_:
        bin_transform(df, bin_transformer_)
        return df, bin_transformer_

    # Binning disabled
    elif not n_discretize_bins:
        return df, bin_transformer_

    # Binning enabled, and binner needs to be fit
    else:
        df, bin_transformer_ = bin_df(
            df,
            n_discretize_bins=n_discretize_bins,
            ignore_feats=ignore_feats,
            verbosity=verbosity,
        )
        return df, bin_transformer_


def bin_df(df, n_discretize_bins=10, ignore_feats=[], verbosity=0):
    """Return df with seemingly continuous features binned, and the bin_transformer or None depending on whether binning occurs."""

    if n_discretize_bins is None:
        return df, None

    isbinned = False
    continuous_feats = find_continuous_feats(
        df, n_discretize_bins=n_discretize_bins, ignore_feats=ignore_feats
    )
    if n_discretize_bins:
        if continuous_feats:
            if verbosity == 1:
                print(f"binning data...\n")
            elif verbosity >= 2:
                print(f"binning features {continuous_feats}...")
            binned_df = df.copy()
            bin_transformer = fit_bins(
                binned_df,
                n_bins=n_discretize_bins,
                output=False,
                ignore_feats=ignore_feats,
                verbosity=verbosity,
            )
            binned_df = bin_transform(binned_df, bin_transformer)
            isbinned = True
    else:
        n_unique_values = sum(
            [len(u) for u in [df[f].unique() for f in continuous_feats]]
        )
        warning_str = f"There are {len(continuous_feats)} features to be treated as continuous: {continuous_feats}. \n Treating {n_unique_values} numeric values as nominal or discrete. To auto-discretize features, assign a value to parameter 'n_discretize_bins.'"
        _warn(warning_str, RuntimeWarning, filename="base", funcname="bin_df")
    if isbinned:
        return binned_df, bin_transformer
    else:
        return df, None


def find_continuous_feats(df, n_discretize_bins, ignore_feats=[]):
    """Return names of df features that seem to be continuous."""

    if not n_discretize_bins:
        return []

    # Find numeric features
    cont_feats = df.select_dtypes(np.number).columns

    # Remove discrete features
    cont_feats = [f for f in cont_feats if len(df[f].unique()) > n_discretize_bins]

    # Remove ignore features
    cont_feats = [f for f in cont_feats if f not in ignore_feats]

    return cont_feats


def fit_bins(df, n_bins=5, output=False, ignore_feats=[], verbosity=0):
    """
    Returns a dict definings fits for numerical features
    A fit is an ordered list of tuples defining each bin's range (min is exclusive; max is inclusive)

    Returned dict allows for fitting to training data and applying the same fit to test data
    to avoid information leak.
    """

    def bin_fit_feat(df, feat, n_bins=10):
        """
        Returns list of tuples defining bin ranges for a numerical feature
        """
        n_bins = min(
            n_bins, len(df[feat].unique())
        )  # In case there are fewer unique values than n_bins
        bin_size = len(df) // n_bins
        sorted_df = df.sort_values(by=[feat])
        sorted_values = sorted_df[feat].tolist()

        if verbosity >= 4:
            print(
                f"{feat}: fitting {len(df[feat].unique())} unique vals into {n_bins} bins"
            )
        bin_ranges = []
        finish_i = -1
        sizes = []

        bin = 0
        start_i = 0
        while bin < n_bins and start_i < len(sorted_values):
            finish_i = min(start_i + bin_size, len(sorted_df) - 1)
            while (
                finish_i < len(sorted_df) - 1
                and finish_i != 0
                and sorted_df.iloc[finish_i][feat] == sorted_df.iloc[finish_i - 1][feat]
            ):  # ensure next bin begins on a new value
                finish_i += 1
            sizes.append(finish_i - start_i)
            if verbosity >= 4:
                print(
                    f"bin #{bin}, start_idx {start_i} value: {sorted_df.iloc[start_i][feat]}, finish_idxx {finish_i} value: {sorted_df.iloc[finish_i][feat]}"
                )
            start_val = sorted_values[start_i]
            finish_val = sorted_values[finish_i]
            bin_range = (start_val, finish_val)
            bin_ranges.append(bin_range)

            start_i = finish_i + 1
            bin += 1

        if verbosity >= 4:
            print(
                f"-bin sizes {sizes}; dataVMR={rnd(np.var(df[feat])/np.mean(df[feat]))}, binVMR={rnd(np.var(sizes)/np.mean(sizes))}"
            )  # , axis=None, dtype=None, out=None, ddof=0)})
        return bin_ranges

    # Create dict to store fit definitions for each feature
    fit_dict = {}
    feats_to_fit = find_continuous_feats(df, n_bins, ignore_feats=ignore_feats)
    if verbosity == 2:
        print(f"fitting bins for features {feats_to_fit}")
    if verbosity >= 2:
        print()

    # Collect fits in dict
    count = 1
    for feat in feats_to_fit:
        fit = bin_fit_feat(df, feat, n_bins=n_bins)
        fit_dict[feat] = fit
    return fit_dict


def bin_transform(df, fit_dict, names_precision=2):
    """
    Uses a pre-collected dictionary of fits to transform df features into bins.
    Returns the fit df rather than modifying inplace.
    """

    def bin_transform_feat(df, feat, bin_fits, names_precision=names_precision):
        """
        Returns new dataframe with n_bin bins replacing each numerical feature
        """

        def renamed(bin_fit_list, value, names_precision=names_precision):
            """
            Returns bin string name for a given numerical value
            Assumes bin_fit_list is ordered
            """
            min_val, min_bin = bin_fit_list[0][0], bin_fit_list[0]
            max_val, max_bin = bin_fit_list[-1][1], bin_fit_list[-1]
            for bin_fit in bin_fits:
                if value <= bin_fit[1]:
                    start_name = str(round(bin_fit[0], names_precision))
                    finish_name = str(round(bin_fit[1], names_precision))
                    bin_name = "-".join([start_name, finish_name])
                    return bin_name
            if value <= min_val:
                return min_bin
            elif value >= max_val:
                return max_bin
            else:
                raise ValueError("No bin found for value", value)

        renamed_values = []
        for value in df[feat]:
            bin_name = renamed(bin_fits, value, names_precision)
            renamed_values.append(bin_name)
        return renamed_values

    # Replace each feature with bin transformations
    for feat, bin_fits in fit_dict.items():
        if feat in df.columns:
            feat_transformation = bin_transform_feat(
                df, feat, bin_fits, names_precision=names_precision
            )
            df[feat] = feat_transformation
    return df


def _try_rename_features(df, class_feat, feature_names):
    """ Rename df columns according to user request. """
    # Rename if same number of features
    df_columns = [col for col in df.columns.tolist() if col != class_feat]
    if len(df_columns) == len(feature_names):
        col_replacements_dict = {
            old: new for old, new in zip(df_columns, feature_names)
        }
        df = df.rename(columns=col_replacements_dict)
        return df
    # Wrong number of feature names
    else:
        return None



import setuptools

setuptools.setup(
    name="wittgenstein",
    version="0.2.0",
    #license="MIT",
    description="Ruleset covering algorithms for explainable machine learning",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Ilan Moscovitz",
    author_email="ilan.moscovitz@gmail.com",
    url="https://github.com/imoscovitz/wittgenstein",
    keywords=[
        "Classification",
        "Decision Rule",
        "Machine Learning",
        "Explainable Machine Learning",
        "Data Science",
    ],
    packages=setuptools.find_packages(),
    install_requires=["pandas", "numpy"],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    include_package_data=True,
    package_data={"": ["data/*.csv"]},
)
